# -*- coding: utf-8 -*-
'''
  fastmat/core/strides.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for cython fastmat striding math library. (types and funcs)

  Author      : wcw
  Introduced  : 2017-11-13
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
import numpy as np
cimport numpy as np

from .types cimport *

cimport numpy as np

################################################## Stride object
ctypedef struct STRIDE_s:
    char *          base
    intsize         strideElement
    intsize         strideVector
    intsize         numElements
    intsize         numVectors
    np.uint8_t      sizeItem
    ftype           dtype

################################################## Basic stride operations
cdef void strideInit(STRIDE_s *, np.ndarray, np.uint8_t)
cdef void strideCopy(STRIDE_s *, STRIDE_s *)
cdef void strideSliceVectors(STRIDE_s *, intsize, intsize, intsize)
cdef void strideSliceElements(STRIDE_s *, intsize, intsize, intsize)
cdef void strideSubgridVector(STRIDE_s *,
                              intsize, intsize,
                              intsize, intsize, intsize, intsize)
cdef void strideFlipVectors(STRIDE_s *)
cdef void strideFlipElements(STRIDE_s *)

cdef stridePrint(STRIDE_s *, text=?)


################################################## Operations with strides
cdef opCopyVector(STRIDE_s *, intsize, STRIDE_s *, intsize)
cdef opZeroVector(STRIDE_s *, intsize)
cdef opZeroVectors(STRIDE_s *)



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/core/cmath.pxd
 -------------------------------------------------- part of the fastmat package

  fastmat math library for cython. (types and funcs)

  Author      : wcw
  Introduced  : 2016-10-07
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
from libc.math cimport ceil

import numpy as np
cimport numpy as np

from .types cimport *
from timeit import timeit

# initialize numpy C-API interface for cython extension-type classes.
# Theis call is required once for every module that uses PyArray_ calls
# WARNING: DO NOT REMOVE THIS LINE OR SEGFAULTS ARE ABOUT TO HAPPEN!
np.import_array()

################################################################################
###  Complexity estimation routines
################################################################################
cdef int _findFFTFactors(int targetLength, int maxFactor,
                         int state, int bestState):
    cdef int ff, length, complexity, newState
    for ff in range(maxFactor, 0, -1):
        length = (state & 0xFFFF) * ff
        complexity = (state >> 16) + ff + 1
        newState = (complexity << 16) + length
        if newState <= bestState and length < targetLength:
            # still now filled targeLength? -> recurse deeper but limit max
            # factor of next stage for speed (still complete)
            bestState = _findFFTFactors(targetLength, ff, newState, bestState)
        else:
            # this iteration ist better than wahat we already found?
            if newState < bestState:
                bestState = newState

            continue

    return bestState


cpdef intsize _findOptimalFFTSize(intsize order, int maxStage):
    cdef intsize paddedSize = 1, minimalSize = order
    cdef float remaining = minimalSize
    cdef int x, complexity, length
    # fill the factor reasonably far with 4-DFTs but leave some room for
    # choosing a close-to-optimal combination of DFT sizes to come as
    # close as reasonable to the targeted size
    while remaining > 64:
        paddedSize *= 4
        remaining /= 4

    x = int(ceil(remaining))

    # for what's left of minimalSize choose the "last mile" optimally, i.e.
    # accept some overshoot while heading for the optimal stage combination
    if x != 1:
        # start a recursion. Pass a default as "current best solution":
        # three stages of 4-DFT, (with complexity of 15 [3 * 4 + 1])
        length = 64
        complexity = 3 * (4 + 1)
        factor = _findFFTFactors(x, maxStage, 1,
                                 (complexity << 16) + length) & 0xFFFF
        paddedSize *= factor

    return paddedSize


cpdef float _getFFTComplexity(intsize N):
    '''
    Return an estimate on the complexity of a typical FFT algorithm.

    Consider a DFT of size N to be pieced together by smaller DFTs corresponding
    to the prime factors of N. Each factor F introduces one level of (N / F)
    DFTs of size F each and a scalar multiplication of the stages' output. The
    composite complexity of all stages is extended by one stage for
    initialization of input and permutation of output.
    '''

    # determine complexity on-the-fly during prime factor decomposition
    # single should be just enough as the largest N is 2^63 which leads to
    # an floating-point exponent of +126 that single still can hold
    cdef intsize nn, factor, x
    cdef float complexity = 0, floatN = N
    cdef ii

    # let nn denote the remaining "non-explored" fraction of N
    nn = N

    # first, consider Radix-4 stages as these are the most efficient and
    # desirable.
    while nn % 4 == 0:
        complexity += 4 + 1
        nn /= 4

    # as 2 * 2 == 4 there can only be one more factor of 2
    if nn > 1 and nn & 1 == 0:
        complexity += 2 + 1
        nn /= 2

    # now consider all odd numbers larger or equal to three as dividers
    # only search up to sqrt(nn), which represents the largest single
    # factor with product nn.
    ii = 3
    while (nn > 1) and (ii * ii < nn):
        if nn % ii == 0:
            complexity += ii + 1
            nn /= ii
        else:
            ii += 2

    # if nn is still not 1 the remainder must be a prime by itself
    # also include the output permutation stage
    if nn > 1:
        complexity += nn + 1

    return floatN * (complexity + 1)


################################################################################
###  Profiling calls
################################################################################


def profileCall(reps, call, *args):
    '''
    wrapper for measuring the runtime of 'call' by averaging the runtime
    of many repeated calls.

        reps        - number of repetitions to be averaged over
        call        - pointer to the evaluatee call
        *args       - a list of parameters for the callee
    '''
    cdef object arg1
    cdef object arg2
    cdef intsize N

    N = 1 if reps < 1 else reps

    def _inner():
        cdef intsize ii
        for _ii in range(N):
            call(*args)

    def _inner1():
        cdef intsize ii
        for _ii in range(N):
            call(arg1)

    def _inner2():
        cdef intsize ii
        for _ii in range(N):
            call(arg1, arg2)

    if len(args) == 1:
        arg1 = args[0]
        runtime = timeit(_inner1, number=1)
    elif len(args) == 2:
        arg1 = args[0]
        arg2 = args[1]
        runtime = timeit(_inner2, number=1)
    else:
        _innerArgs = args
        runtime = timeit(_inner, number=1)

    # return results
    return {
        'avg': runtime / reps,
        'time': runtime,
        'cnt': reps
    }


################################################################################
###  Array creation routines
################################################################################

################################################## _arrZero()
cpdef np.ndarray _arrZero(
    int dims,
    intsize numN,
    intsize numM,
    nptype dtype,
    bint fortranStyle=True
):
    '''
    Create and zero-init new ndarray of specified shape and data type
    (up to two dimensions).
    '''
    cdef np.npy_intp shape[2]
    shape[0] = numN
    shape[1] = numM

    return np.PyArray_ZEROS(
        dims if dims < 2 else 2,    # Nr. Dimensions
        & shape[0],                 # Sizes of Dimensions
        dtype,                      # Data Type of elements
        fortranStyle                # FORTRAN-style result
    )


################################################## _arrEmpty()
cpdef np.ndarray _arrEmpty(
    int dims,
    intsize numN,
    intsize numM,
    nptype dtype,
    bint fortranStyle=True
):
    '''
    Create an empty ndarray of specified shape and data type
    (up to two dimensions)
    '''
    cdef np.npy_intp shape[2]
    shape[0] = numN
    shape[1] = numM

    return np.PyArray_EMPTY(
        dims if dims < 2 else 2,    # Nr. Dimensions
        & shape[0],                 # Sizes of Dimensions
        dtype,                      # Data Type of elements
        fortranStyle                # FORTRAN-style result
    )


cpdef np.ndarray _arrReshape(
    np.ndarray arr,
    int dims,
    intsize numN,
    intsize numM,
    np.NPY_ORDER order
):
    cdef np.PyArray_Dims shape2D
    cdef np.npy_intp shape[2]
    shape2D.ptr = &shape[0]
    shape2D.len = dims if dims < 2 else 2
    shape[0] = numN
    shape[1] = numM

    return np.PyArray_Newshape(arr, &shape2D, order)


cpdef bint _arrResize(
    np.ndarray arr,
    int dims,
    intsize numN,
    intsize numM,
    np.NPY_ORDER order
):
    cdef np.PyArray_Dims shape2D
    cdef np.npy_intp shape[2]
    shape2D.ptr = &shape[0]
    shape2D.len = dims if dims < 2 else 2
    shape[0] = numN
    shape[1] = numM

    return np.PyArray_Resize(arr, &shape2D, False, order) is None


cpdef np.ndarray _arrCopyExt(
    np.ndarray arr,
    nptype dtype,
    int flags
):
    return np.PyArray_FROM_OTF(arr, dtype, flags)


cpdef np.ndarray _arrForceType(
    np.ndarray arr,
    nptype typeArr
):
    return arr if (np.PyArray_TYPE(arr) == typeArr) \
        else np.PyArray_FROM_OT(arr, typeArr)


cpdef np.ndarray _arrForceAlignment(
    np.ndarray arr,
    int flags,
    bint fortranStyle=True
):
    if np.PyArray_ISONESEGMENT(arr) and \
       (np.PyArray_ISFORTRAN(arr) == fortranStyle) and \
       np.PyArray_ISCONTIGUOUS(arr):
        return arr

    if fortranStyle:
        flags += np.NPY_F_CONTIGUOUS
    else:
        flags += np.NPY_C_CONTIGUOUS

    return np.PyArray_FROM_OF(arr, np.NPY_OWNDATA + np.NPY_ENSUREARRAY + flags)


cpdef np.ndarray _arrForceTypeAlignment(
    np.ndarray arr,
    nptype typeArr,
    int flags,
    bint fortranStyle=True
):
    if (np.PyArray_TYPE(arr) == typeArr) and \
            np.PyArray_ISONESEGMENT(arr) and \
            (np.PyArray_ISFORTRAN(arr) == fortranStyle) and \
            np.PyArray_ISCONTIGUOUS(arr):
        return arr

    if fortranStyle:
        flags += np.NPY_F_CONTIGUOUS
    else:
        flags += np.NPY_C_CONTIGUOUS

    return np.PyArray_FROM_OTF(arr, typeArr,
                               np.NPY_OWNDATA + np.NPY_ENSUREARRAY + flags)


################################################## fused-type conjugate
cdef void _conjugateMV1(
    TYPE_COMPLEX[:] input,
    TYPE_COMPLEX[:] output
):
    '''
    Conjugate contents of memoryview in input and write to output.
    If the memoryview data type is not complex return False, indicating
    no processing of input and changes to output.
    '''
    cdef intsize nn, N = input.shape[0]
    # conjugate elements from input, write to output
    for nn in range(N):
        output[nn].real = input[nn].real
        output[nn].imag = -input[nn].imag


cdef void _conjugateMV2(
    TYPE_COMPLEX[:, :] input,
    TYPE_COMPLEX[:, :] output
):
    '''
    Conjugate contents of memoryview in input and write to output.
    If the memoryview data type is not complex return False, indicating
    no processing of input and changes to output.
    '''
    cdef intsize nn, mm, N = input.shape[0], M = input.shape[1]
    # conjugate elements from input, write to output
    for mm in range(M):
        for nn in range(N):
            output[nn, mm].real = input[nn, mm].real
            output[nn, mm].imag = -input[nn, mm].imag


cpdef np.ndarray _conjugate(np.ndarray arr):
    '''
    Conjugate numpy ndarray non-destructively.
    Return a conjugated copy if complex input or original data view if real.
    '''
    # determine type of array
    typeArr = typeSelection[np.PyArray_TYPE(arr)]
    if typeArr == TYPE_INVALID:
        raise NotImplementedError("Type %d not supported." % (typeArr))

    # check if complex. If not, abort
    if not ((typeArr == TYPE_COMPLEX64) or (typeArr == TYPE_COMPLEX128)):
        return arr

    # check if unsupported data type
    if typeArr == TYPE_INVALID:
        raise NotImplementedError("Type %d not supported." % (typeArr))

    # create empty output ndarray with equal shape as input (FORTRAN-style)
    cdef int numDims = arr.ndim
    cdef np.ndarray arrOutput = _arrEmpty(
        numDims,
        arr.shape[0],
        arr.shape[1] if numDims > 1 else 1,
        typeInfo[typeArr].typeNum
    )

    # perform conjugation
    if numDims == 1:
        if typeArr == TYPE_COMPLEX64:
            _conjugateMV1[np.complex64_t](arr, arrOutput)
        elif typeArr == TYPE_COMPLEX128:
            _conjugateMV1[np.complex128_t](arr, arrOutput)
    elif numDims == 2:
        if typeArr == TYPE_COMPLEX64:
            _conjugateMV2[np.complex64_t](arr, arrOutput)
        elif typeArr == TYPE_COMPLEX128:
            _conjugateMV2[np.complex128_t](arr, arrOutput)

    return arrOutput


################################################## in-place fused-type conjugate

cdef void _conjInplaceCore(
    np.ndarray arr,
    TYPE_COMPLEX typeArr
):
    '''
    Conjugate a raw data buffer of type TYPE_COMPLEX and cnt elements,
    assuming to be consecutive in memory
    '''
    cdef intsize ii, cnt = np.PyArray_SIZE(arr)
    cdef TYPE_COMPLEX * pData = <TYPE_COMPLEX * > arr.data

    # conjugate elements from input, write to output
    for ii in range(cnt):
        pData[ii].imag = -pData[ii].imag


cpdef bint _conjugateInplace(np.ndarray arr):
    '''
    Conjugate numpy ndarray in input and write to output.
    Performs as a wrapper for ndarrays
    '''
    # determine type of array
    cdef nptype typeArr = _getFType(arr)

    # check if complex. If so, conjugate.
    if typeArr == TYPE_COMPLEX64:
        _conjInplaceCore[np.complex64_t](arr, typeArr)
        return True
    elif typeArr == TYPE_COMPLEX128:
        _conjInplaceCore[np.complex128_t](arr, typeArr)
        return True
    elif typeArr == TYPE_INVALID:
        raise NotImplementedError("Type %d not supported." % (typeArr))

    return False


################################################## in-place fused-type conjugate
cdef np.float64_t _norm(
    TYPE_ALL *vec,
    intsize N
):
    '''Compute the norm of the vector `vec`.'''
    cdef intsize nn
    cdef TYPE_ALL val
    cdef np.float64_t norm = 0.

    for nn in range(N):
        val = vec[nn]
        if (TYPE_ALL == np.complex64_t) or (TYPE_ALL == np.complex128_t):
            norm += val.real * val.real
            norm += val.imag * val.imag
        else:
            norm += val * val

    return norm


cdef np.float64_t _normMV(TYPE_ALL[:] vec):
    '''Compute the norm of the vector `vec`.'''
    cdef intsize nn, N = vec.shape[0]
    cdef TYPE_ALL val
    cdef np.float64_t norm = 0.

    for nn in range(N):
        val = vec[nn]
        if (TYPE_ALL == np.complex64_t) or (TYPE_ALL == np.complex128_t):
            norm += val.real * val.real + val.imag * val.imag
        else:
            norm += val * val

    return norm


cdef TYPE_ALL _corrMV(
    TYPE_ALL[:] vec1,
    TYPE_ALL[:] vec2
):
    '''Compute the correlation of the vectors `vec1` and `vec2`.'''
    cdef intsize nn, N = vec1.shape[0]
    cdef TYPE_ALL corr = 0

    for nn in range(N):
        corr = vec1[nn] * vec2[nn]

    return corr


################################################## fused-type multiply
cdef void _opCoreI(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    TYPE_IN_I tIn,
    TYPE_OP_I tOp,
    TYPE_INT tOut,
    OP_MODE mode,
    intsize param
):
    '''
    '''
    # never use arrIn.shape[1] without checking dimension count of arrIn !
    cdef intsize kk, nn, mm
    cdef intsize N = arrIn.shape[0]
    cdef intsize NO = arrOut.shape[0], MO = arrOut.shape[1]

    # pointers to data field, assume ordering as contiguous columns
    cdef TYPE_IN_I * pIn = <TYPE_IN_I * > arrIn.data
    cdef TYPE_OP_I * pOp = <TYPE_OP_I * > arrOp.data
    cdef TYPE_INT * pOut = <TYPE_INT * > arrOut.data

    # pointers to a single column (arrOptor)
    cdef TYPE_IN_I * pInVec
    cdef TYPE_OP_I * pOpVec
    cdef TYPE_INT * pOutVec

    # memoryviews
    cdef TYPE_OP_I[:] mvOp

    if mode == MODE_MUL:
        # element-wise multiplication of arrIn and arrOp in m-Dimension
        # arrOp is assumed to be a one-dimensional vector
        for mm in range(MO):
            # possible issue with strides for non-aligned data
            pInVec = &(pIn[N * mm])
            pOutVec = &(pOut[N * mm])
            for nn in range(N):
                pOutVec[nn] = pInVec[nn] * pOp[nn]
    elif mode == MODE_DOTROW:
        # matrix dot-product or arrOp and arrIn.

        if (arrOp.ndim > 1) or (arrOp.shape[0] != N):
            raise ValueError("Dimension mismatch in _dotRowOffset")

        # iterate over columns of input / output data
        mvOp = arrOp
        for mm in range(MO):
            pOutVec = &(pOut[NO * mm + param])
            pInVec = &(pIn[N * mm])

            pOutVec[0] = pInVec[0] * mvOp[0]
            for nn in range(1, N):
                pOutVec[0] += pInVec[nn] * mvOp[nn]


cdef void _opCoreF(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    TYPE_IN_R tIn,
    TYPE_OP_R tOp,
    TYPE_REAL tOut,
    OP_MODE mode,
    intsize param
):
    '''
    '''
    # never use arrIn.shape[1] without checking dimension count of arrIn !
    cdef intsize kk, nn, mm
    cdef intsize N = arrIn.shape[0]
    cdef intsize NO = arrOut.shape[0], MO = arrOut.shape[1]

    # pointers to data field, assume ordering as contiguous columns
    cdef TYPE_IN_R * pIn = <TYPE_IN_R * > arrIn.data
    cdef TYPE_OP_R * pOp = <TYPE_OP_R * > arrOp.data
    cdef TYPE_REAL * pOut = <TYPE_REAL * > arrOut.data

    # pointers to a single column (arrOptor)
    cdef TYPE_IN_R * pInVec
    cdef TYPE_OP_R * pOpVec
    cdef TYPE_REAL * pOutVec

    # memoryviews
    cdef TYPE_OP_R[:] mvOp

    if mode == MODE_MUL:
        # element-wise multiplication of arrIn and arrOp in m-Dimension
        # arrOp is assumed to be a one-dimensional vector
        for mm in range(MO):
            # possible issue with strides for non-aligned data
            pInVec = &(pIn[N * mm])
            pOutVec = &(pOut[N * mm])
            for nn in range(N):
                pOutVec[nn] = pInVec[nn] * pOp[nn]
    elif mode == MODE_DOTROW:
        # matrix dot-product or arrOp and arrIn.

        if (arrOp.ndim > 1) or (arrOp.shape[0] != N):
            raise ValueError("Dimension mismatch in _dotRowOffset")

        # iterate over columns of input / output data
        mvOp = arrOp
        for mm in range(MO):
            pOutVec = &(pOut[NO * mm + param])
            pInVec = &(pIn[N * mm])

            pOutVec[0] = pInVec[0] * mvOp[0]
            for nn in range(1, N):
                pOutVec[0] += pInVec[nn] * mvOp[nn]


cdef void _opCoreC(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    TYPE_IN tIn,
    TYPE_OP tOp,
    TYPE_COMPLEX tOut,
    OP_MODE mode,
    intsize param
):
    '''
    '''
    # never use arrIn.shape[1] without checking dimension count of arrIn !
    cdef intsize kk, nn, mm
    cdef intsize N = arrIn.shape[0]
    cdef intsize NO = arrOut.shape[0], MO = arrOut.shape[1]

    # pointers to data field, assume ordering as contiguous columns
    cdef TYPE_IN * pIn = <TYPE_IN * > arrIn.data
    cdef TYPE_OP * pOp = <TYPE_OP * > arrOp.data
    cdef TYPE_COMPLEX * pOut = <TYPE_COMPLEX * > arrOut.data

    # pointers to a single column (arrOptor)
    cdef TYPE_IN * pInVec
    cdef TYPE_OP * pOpVec
    cdef TYPE_COMPLEX * pOutVec

    # memoryviews
    cdef TYPE_OP[:] mvOp

    if mode == MODE_MUL:
        # element-wise multiplication of arrIn and arrOp in m-Dimension
        # arrOp is assumed to be a one-dimensional vector
        for mm in range(MO):
            # possible issue with strides for non-aligned data
            pInVec = &(pIn[N * mm])
            pOutVec = &(pOut[N * mm])
            for nn in range(N):
                pOutVec[nn] = pInVec[nn] * pOp[nn]
    elif mode == MODE_DOTROW:
        # matrix dot-product or arrOp and arrIn.

        if (arrOp.ndim > 1) or (arrOp.shape[0] != N):
            raise ValueError("Dimension mismatch in _dotRowOffset")

        # iterate over columns of input / output data
        mvOp = arrOp
        for mm in range(MO):
            pOutVec = &(pOut[NO * mm + param])
            pInVec = &(pIn[N * mm])

            pOutVec[0] = pInVec[0] * mvOp[0]
            for nn in range(1, N):
                pOutVec[0] = pOutVec[0] + pInVec[nn] * mvOp[nn]


################################ dispatch integer results
cdef void _opI(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    TYPE_INT tOut,
    OP_MODE mode,
    intsize param
):
    '''    '''
    # dispatch specialization of core routine according tOptor
    if tIn == TYPE_INT8:
        if tOp == TYPE_INT8:
            _opCoreI[np.int8_t, np.int8_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT32:
            _opCoreI[np.int8_t, np.int32_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT64:
            _opCoreI[np.int8_t, np.int64_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT32:
        if tOp == TYPE_INT8:
            _opCoreI[np.int32_t, np.int8_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT32:
            _opCoreI[np.int32_t, np.int32_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT64:
            _opCoreI[np.int32_t, np.int64_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT64:
        if tOp == TYPE_INT8:
            _opCoreI[np.int64_t, np.int8_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT32:
            _opCoreI[np.int64_t, np.int32_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
        elif tOp == TYPE_INT64:
            _opCoreI[np.int64_t, np.int64_t, TYPE_INT](
                arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)


################################ dispatch float results
cdef void _opRIn(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    TYPE_OP_R tOp,
    TYPE_REAL tOut,
    OP_MODE mode,
    intsize param
):
    '''     '''
    # dispatch specialization of core routine according tOptor
    if tIn == TYPE_INT8:
        _opCoreF[np.int8_t, TYPE_OP_R, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT32:
        _opCoreF[np.int32_t, TYPE_OP_R, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT64:
        _opCoreF[np.int64_t, TYPE_OP_R, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_FLOAT32:
        _opCoreF[np.float32_t, TYPE_OP_R, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_FLOAT64:
        _opCoreF[np.float64_t, TYPE_OP_R, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)


cdef void _opR(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    TYPE_REAL tOut,
    OP_MODE mode,
    intsize param
):
    '''    '''
    # dispatch specialization of core routine according tOptor
    if tOp == TYPE_INT8:
        _opRIn[np.int8_t, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_INT32:
        _opRIn[np.int32_t, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_INT64:
        _opRIn[np.int64_t, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_FLOAT32:
        _opRIn[np.float32_t, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_FLOAT64:
        _opRIn[np.float64_t, TYPE_REAL](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)


################################ dispatch complex results
cdef void _opCIn(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    TYPE_OP tOp,
    TYPE_COMPLEX tOut,
    OP_MODE mode,
    intsize param
):
    '''    '''
    # dispatch specialization of core routine according tOptor
    if tIn == TYPE_INT8:
        _opCoreC[np.int8_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT32:
        _opCoreC[np.int32_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_INT64:
        _opCoreC[np.int64_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_FLOAT32:
        _opCoreC[np.float32_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_FLOAT64:
        _opCoreC[np.float64_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_COMPLEX64:
        _opCoreC[np.complex64_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tIn == TYPE_COMPLEX128:
        _opCoreC[np.complex128_t, TYPE_OP, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)


cdef void _opC(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    TYPE_COMPLEX tOut,
    OP_MODE mode,
    intsize param
):
    '''    '''
    # dispatch specialization of core routine according tOptor
    if tOp == TYPE_INT8:
        _opCIn[np.int8_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_INT32:
        _opCIn[np.int32_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_INT64:
        _opCIn[np.int64_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_FLOAT32:
        _opCIn[np.float32_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_FLOAT64:
        _opCIn[np.float64_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_COMPLEX64:
        _opCIn[np.complex64_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOp == TYPE_COMPLEX128:
        _opCIn[np.complex128_t, TYPE_COMPLEX](
            arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)


cdef void _op(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    ftype tOut,
    OP_MODE mode,
    intsize param
):
    # dispatch specialization of core routines according typeData
    if tOut == TYPE_INT8:
        _opI[np.int8_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_INT32:
        _opI[np.int32_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_INT64:
        _opI[np.int64_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_FLOAT32:
        _opR[np.float32_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_FLOAT64:
        _opR[np.float64_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_COMPLEX64:
        _opC[np.complex64_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)
    elif tOut == TYPE_COMPLEX128:
        _opC[np.complex128_t](arrIn, arrOp, arrOut, tIn, tOp, tOut, mode, param)

################################ entry point: _multiply()

cpdef _multiply(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    ftype tOut
):
    '''
    '''
    _op(arrIn, arrOp, arrOut, tIn, tOp, tOut, MODE_MUL, 0)


################################ entry point: _dotRow()

cpdef _dotSingleRow(
    np.ndarray arrIn,
    np.ndarray arrOp,
    np.ndarray arrOut,
    ftype tIn,
    ftype tOp,
    ftype tOut,
    intsize iRow
):
    _op(arrIn, arrOp, arrOut, tIn, tOp, tOut, MODE_DOTROW, iRow)



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=True
'''
  fastmat/MLToeplitz.py
 -------------------------------------------------- part of the fastmat package

  MLToeplitz matrix.


  Author      : sempersn
  Introduced  : 2017-09-19
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .core.cmath cimport *
from .Matrix cimport Matrix
from .Partial cimport Partial
from .Product cimport Product
from .Fourier cimport Fourier
from .Diag cimport Diag
from .Kron cimport Kron
from .Toeplitz cimport Toeplitz


################################################################################
################################################## class Toeplitz
cdef class MLToeplitz(Partial):

    ############################################## class properties
    # vecC - Property (read-only)
    # Return the matrix-defining column vector of the circulant matrix
    property tenT:
        def __get__(self):
            return self._tenT

    ############################################## class methods
    def __init__(self, tenT, **options):
        '''
        Initialize MLToeplitz Matrix instance.
        '''

        self._tenT = np.atleast_1d(np.squeeze(np.copy(tenT)))

        assert self._tenT.ndim >= 1, "Column-definition vector must be >= 1D."

        # extract the level dimensions from the defining tensor
        self._arrN = (
            (np.array(np.array(self._tenT).shape) + 1) /2).astype('int')

        # get the size of the matrix
        numN = np.prod(self._arrN)

        # stages during optimization
        cdef int maxStage = options.get('maxStage', 4)

        # minimum number to pad to during optimization and helper arrays
        arrNpad = 2 * self._arrN - 1
        arrOptSize = np.zeros_like(self._arrN)
        arrDoOpt = np.zeros_like(self._arrN)

        cdef bint optimize = options.get('optimize', True)
        if optimize:
            # go through all level dimensions and get optimal FFT size
            for inn, nn in enumerate(arrNpad):
                arrOptSize[inn] = _findOptimalFFTSize(nn, maxStage)

                # use this size, if we get better in that level
                if _getFFTComplexity(arrOptSize[inn]) < _getFFTComplexity(nn):
                    arrDoOpt[inn] = 1

        arrDoOpt = arrDoOpt == 1
        arrNopt = np.copy(2 * self._arrN - 1)

        # set the optimization size to the calculated one
        arrNopt[arrDoOpt] = arrOptSize[arrDoOpt]

        # get the size of the zero padded matrix
        numNopt = np.prod(arrNopt)

        # allocate memory for the tensor in MD fourier domain
        tenThat = np.empty_like(self._tenT, dtype='complex')
        tenThat[:] = self._tenT[:]

        # go through the array and apply the preprocessing in direction
        # of each axis. this cannot be done without the for loop, since
        # manipulations always influence the data for the next dimension
        for ii in range(len(self._arrN)):
            tenThat = np.apply_along_axis(
                self._preProcSlice,
                ii,
                tenThat,
                ii,
                arrNopt,
                self._arrN
            )

        # after correct zeropadding, go into fourier domain
        tenThat = np.fft.fftn(tenThat).reshape(numNopt) / numNopt

        # subselection array
        arrIndices = np.arange(numNopt)[self._genArrS(self._arrN, arrNopt)]

        # create the decomposing kronecker product
        cdef Kron KN = Kron(*list(map(
            lambda ii : Fourier(ii, optimize=False), arrNopt
        )))

        # now decompose the ML matrix as a product
        cdef Product P = Product(KN.H, Diag(tenThat),
                                 KN, **options)

        # initialize Partial of Product. Only use Partial when padding size
        if np.allclose(self._arrN, arrNopt):
            super(MLToeplitz, self).__init__(P)
        else:
            super(MLToeplitz, self).__init__(P, N=arrIndices, M=arrIndices)

        # Currently Fourier matrices bloat everything up to complex double
        # precision, therefore make sure tenT matches the precision of the
        # matrix itself
        if self.dtype != self._tenT.dtype:
            self._tenT = self._tenT.astype(self.dtype)

    cpdef np.ndarray _getArray(self):
        '''Return an explicit representation of the matrix as numpy-array.'''
        return self._reference()

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (0., 0.)

    def _preProcSlice(
        self,
        theSlice,
        numSliceInd,
        arrNopt,
        arrN
    ):
        '''
        preprocess one axis of the defining tensor. here we check for one
        dimension, whether it makes sense to  zero-pad or not by estimating
        the fft-complexity in each dimension.
        '''
        arrRes = np.empty(1)

        if arrNopt[numSliceInd] > 2 *arrN[numSliceInd] -1:
            z = np.zeros(arrNopt[numSliceInd] - 2 *arrN[numSliceInd] +1)
            arrRes = np.concatenate((
                np.copy(theSlice[:arrN[numSliceInd]]),
                z,
                np.copy(theSlice[arrN[numSliceInd]:])
            ))
        else:
            arrRes = np.copy(theSlice)

        return arrRes

    def _genArrS(
        self,
        arrN,
        arrNout,
        verbose=False
    ):
        '''
        Iteratively filter out the non-zero elements in the padded version
        of X. I know, that one can achieve this from a zero-padded
        version of the tensor Xpadten, but the procedure itself is very
        helpful for understanding how the nested levels have an impact on
        the padding structure
        '''
        n = arrN.shape[0]
        numNout = np.prod(arrNout)
        arrS = np.arange(numNout) >= 0
        for ii in range(n):
            if verbose:
                print("state", arrS)
                print("modulus", np.mod(
                    np.arange(numNout),
                    np.prod(arrNout[:(n -ii)])
                ))
                print("inequ", arrN[n -1 -ii] * np.prod(arrNout[:(n -1 -ii)]))
                print("res", np.mod(
                    np.arange(numNout),
                    np.prod(arrNout[:(n -1 -ii)])
                ) < arrN[n -1 -ii] * np.prod(arrNout[:(n -1 -ii)]))
            # iteratively subselect more and more indices in arrS
            np.logical_and(
                arrS,
                np.mod(
                    np.arange(numNout),
                    np.prod(arrNout[ii:])
                ) < arrN[ii] * np.prod(arrNout[ii +1:]),
                arrS
            )
        return arrS

    cpdef Matrix _getNormalized(self):
        arrNorms = self._normalizeCore(self._tenT)

        return self * Diag(1. / np.sqrt(arrNorms))

    def _normalizeCore(self, tenT):
        cdef intsize ii, numS1, numS2, numS3

        cdef intsize numL = int((tenT.shape[0] + 1) /2)
        cdef intsize numEll = tenT.shape[0]

        cdef intsize numD = len(tenT.shape)

        cdef np.ndarray arrT, arrNorms
        if numD == 1:
            # if we are deep enough we do the normal toeplitz stuff
            arrT = tenT

            arrNorms = np.zeros(numL)

            arrNorms[0] = np.linalg.norm(arrT[:numL]) **2

            for ii in range(numL - 1):

                arrNorms[ii + 1] = arrNorms[ii] \
                    + np.abs(arrT[2 * numL - 2 - ii]) ** 2 \
                    - np.abs(arrT[numL - ii - 1]) ** 2

        else:
            numS1 = np.prod(self._arrN[-numD :])
            numS2 = np.prod(self._arrN[-(numD - 1) :])
            arrNorms = np.zeros(numS1)
            arrT = np.zeros((numEll, numS2))

            # go deeper in recursion and get norms of blocks
            for ii in range(numEll):
                arrT[ii, :] = self._normalizeCore(tenT[ii, :])

            numS3 = arrT.shape[1]
            arrNorms[:numS3] = np.sum(arrT[:numL, :], axis=0)

            # now do blockwise subtraction and addition
            for ii in range(numL - 1):

                arrNorms[
                    (ii +1) *numS2 : (ii +2) *numS2
                ] = arrNorms[ii *numS2 : (ii +1) *numS2] + \
                    + arrT[2 * numL - 2 - ii] \
                    - arrT[numL - ii - 1]

        return arrNorms

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''

        return self._refRecursion(self._arrN, self._tenT, False)

    def _refRecursion(
        self,
        arrN,               # dimensions in each level
        tenU,               # defining elements
        verbose=False       # verbosity flag
    ):
        '''
        Construct a multilevel toeplitz matrix
        '''
        # number of dimensions
        numD = arrN.shape[0]

        # get size of resulting block toeplitz matrix
        numN = np.prod(arrN)

        # get an array of all partial sequential products
        # starting at the front
        arrNprod = np.array(
            list(map(lambda ii : np.prod(arrN[ii:]), range(numN - 1)))
        )

        # permutation array for block placement, since we need to place the
        # blocks in the same fashion, we arrange the elements in the blocks,
        # such that the preprocessing does the right thing
        arrP = np.arange(2 * arrN[0] - 1)
        arrP = np.concatenate((arrP[:arrN[0]][::-1], arrP[arrN[0]:][::-1]))
        arrP = np.argsort(arrP)

        # allocate memory for the result
        T = np.zeros((numN, numN), dtype=self.dtype)

        # check if we can go a least a level deeper
        if numD > 1:

            # iterate over size of the first dimension
            for nn_ in range(2 * arrN[0] - 1):

                # select the right block position with the permutation array
                nn       = arrP[nn_]
                tmp      = nn - arrN[0] + 1
                countAbs = arrN[0] - abs(tmp)
                countSig = 0 if tmp == 0 else tmp / abs(tmp)

                # now calculate the block recursively
                subT     = self._refRecursion(arrN[1 :], tenU[nn_])

                # decide whether it is below or above the diagonal or on it
                # and the act accordingly
                if countSig == 0:
                    # we are on the diagonal
                    for mm in range(countAbs):
                        T[
                            mm * arrNprod[1] : (mm + 1) * arrNprod[1],
                            mm * arrNprod[1] : (mm + 1) * arrNprod[1]
                        ] = subT
                elif countSig < 0:
                    # we are below the diagonal
                    for mm in range(countAbs):
                        T[
                            (mm + abs(tmp))     * arrNprod[1] :
                            (mm + 1 + abs(tmp)) * arrNprod[1],
                            mm       * arrNprod[1]            :
                            (mm + 1) * arrNprod[1]
                        ] = subT
                else:
                    # we are above the diagonal
                    for mm in range(countAbs):
                        T[
                            mm * arrNprod[1] : (mm + 1) * arrNprod[1],
                            (mm + abs(tmp))     * arrNprod[1] :
                            (mm + 1 + abs(tmp)) * arrNprod[1],
                        ] = subT

            return T
        else:
            # if we are in a lowest level, we just construct the right
            # single level toeplitz block
            return Toeplitz(tenU[:numN], tenU[numN:][::-1]).array

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # 35 is just any number that causes no padding
                # 41 is the first size for which bluestein is faster
                TEST.NUM_N      : 27,
                TEST.NUM_M      : TEST.NUM_N,
                'mTypeC'        : TEST.Permutation(TEST.ALLTYPES),
                'optimize'      : True,
                TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                'vecC'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeC',
                    TEST.SHAPE  : (5, 5, 5),
                    TEST.ALIGN  : TEST.PARAMALIGN
                }),
                TEST.INITARGS   : (lambda param : [param['vecC']()]),
                TEST.INITKWARGS : {'optimize' : 'optimize'},
                TEST.OBJECT     : MLToeplitz,
                TEST.NAMINGARGS : dynFormat("%s,optimize=%s",
                                            'vecC', str('optimize')),
                TEST.TOL_POWER  : 2.,
                TEST.TOL_MINEPS : _getTypeEps(np.float64)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c:
                                   MLToeplitz(np.random.randn(
                                       *(2 * [(2 * c - 1) + 2])
                                   ))),
                BENCH.FUNC_SIZE : (lambda c: (c + 1) ** 2),
                BENCH.FUNC_STEP : (lambda c: c * 10 ** (1. / 12))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {},
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: MLToeplitz(
                    np.random.randn(*(2 * [(2 * c - 1) + 2])).astype(dt)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Multilevel Toeplitz Matrix (\texttt{fastmat.MLToeplitz})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
Let $d \geqslant 2$. Then, given a $d$-dimensional complex sequence
$\bm{t} = [t_{\bm{k}}]$ for $\bm{k} \in \N^d$ a
$d$-level Toeplitz matrix $\bm{T}_{\bm{n},d}$ is recursively defined as
%
\[\bm{T}_{(\bm{n},d)} =
\begingroup
\setlength\arraycolsep{3pt}
\begin{bmatrix}
    \bm{T}_{(1,\bm{m}),\ell}        & \bm{T}_{(2 n_1 - 1,\bm{m}),\ell}
    & \dots     & \bm{T}_{(n_1 + 1,\bm{m}),\ell}    \\
    \bm{T}_{(2,\bm{m}),\ell}        & \bm{T}_{(1,\bm{m}),\ell}
    & \dots     & \bm{T}_{(n_1 + 2,\bm{m}),\ell}    \\
    \vdots                          & \vdots
    & \ddots    & \vdots                            \\
    \bm{T}_{(n_1,\bm{m}),\ell}      & \bm{T}_{(n_1 - 1,\bm{m}),\ell}
    & \dots     & \bm{T}_{(1,\bm{m}),\ell}          \\
\end{bmatrix},
\endgroup
\]
%
where $\bm m = \bm n_{-1}$ and $\ell = d-1$. So for $\bm n = [2,2]$ and
$\bm t \in \C^{3 \times 3}$ we get
%
\[
\begingroup
\setlength\arraycolsep{3pt}
\bm T_{[2,2],2} =
\begin{bmatrix}
\bm T_{[1,2],1} & \bm T_{[3,2],1} \\
\bm T_{[2,2],1} & \bm T_{[1,2],1}
\end{bmatrix}
=
\begin{bmatrix}
t_{1,1} & t_{1,3} & t_{3,1} & t_{3,3} \\
t_{1,2} & t_{1,1} & t_{3,2} & t_{3,1} \\
t_{2,1} & t_{2,2} & t_{1,1} & t_{1,3} \\
t_{2,2} & t_{2,1} & t_{1,2} & t_{1,1}
\end{bmatrix}.
\endgroup
\]
                """,
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# construct the',
                            '# parameters',
                            'n = 2',
                            'l = 2',
                            't = np.arange(',
                            '(2 * n - 1) ** l',
                            ').reshape(',
                            '(2 * n - 1, 2 * n - 1)',
                            ')',
                            '',
                            '# construct the matrix',
                            'T = fm.MLToeplitz(t)',
                            caption=r"""
This yields
\[\bm t = \begin{bmatrix}1 & 2 & 3 \\ 4 & 5 & 6 \\ 7 & 8 & 9 \end{bmatrix}\]
\[\bm T = \left(\begin{array}{cccc}
    1 & 3 & 7 & 9 \\
    2 & 1 & 8 & 7 \\
    4 & 5 & 1 & 3 \\
    5 & 4 & 2 & 1
\end{array}\right)\]"""),
                r"""
This class depends on \texttt{Fourier}, \texttt{Diag}, \texttt{Kron},
\texttt{Product} and \texttt{Partial}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm{\mathcal{C}} \in \R^{n \times n}$ with $n \in \N$ and all
entries drawn from a  standard Gaussian distribution.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
'''
  setup.py
 -------------------------------------------------- part of the fastmat package

  Setup script for installation of fastmat package

  Usecases:
    - install fastmat package system-wide on your machine (needs su privileges)
        EXAMPLE:        'python setup.py install'

    - install fastmat package for your local user only (no privileges needed)
        EXAMPLE:        'python setup.py install --user'

    - compile all cython source files locally
        EXAMPLE:        'python setup.py build_ext --inplace'


  Author      : sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

# import modules
import platform
import sys
import os
import re
import subprocess

packageName     = 'fastmat'
packageVersion  = '0.1.1b1'       # provide a version tag as fallback
fullVersion     = packageVersion
strVersionFile  = "%s/version.py" %(packageName)

VERSION_PY      = """
# -*- coding: utf-8 -*-
# This file carries the module's version information which will be updated
# during execution of the installation script, setup.py. Distribution tarballs
# contain a pre-generated copy of this file.

__version__ = '%s'
"""


def WARNING(string):
    print("\033[91m WARNING:\033[0m %s" % (string))


###############################################################################
###  Import setuptools
###############################################################################

# load setup and extensions from setuptools. If that fails, try distutils
try:
    from setuptools import setup, Extension
except ImportError:
    WARNING("Could not import setuptools.")
    raise


###############################################################################
###  Determine package version
###############################################################################
def getCurrentVersion():
    global packageVersion
    global fullVersion

    # check if there is a manual version override
    if os.path.isfile(".version"):
        with open(".version", "r") as f:
            stdout = f.read().split('\n')[0]
        print("Override of version string to '%s' (from .version file )" %(
            stdout))

        packageVersion = stdout

    else:
        # check if source directory is a git repository
        if not os.path.exists(".git"):
            print(("Installing from something other than a Git repository; " +
                   "Version file '%s' untouched.") %(strVersionFile))
            return

        # fetch current tag and commit description from git
        try:
            p = subprocess.Popen(
                ["git", "describe", "--tags", "--dirty", "--always"],
                stdout=subprocess.PIPE
            )
        except EnvironmentError:
            print("Not a git repository; Version file '%s' not touched." %(
                strVersionFile))
            return

        stdout = p.communicate()[0].strip()
        if stdout is not str:
            stdout = stdout.decode()

        if p.returncode != 0:
            print(("Unable to fetch version from git repository; " +
                   "leaving version file '%s' untouched.") %(strVersionFile))
            return

        # output results to version string, extract package version number
        # from git tag
        fullVersion = stdout
        versionMatch = re.match("[.+\d+]+\d*[abr]\d*", fullVersion)
        if versionMatch:
            packageVersion = versionMatch.group(0)
            print("Fetched package version number from git tag (%s)." %(
                packageVersion))


# get version from git and update fastmat/__init__.py accordingly
getCurrentVersion()

# make sure there exists a version.py file in the project
with open(strVersionFile, "w") as f:
    f.write(VERSION_PY % (fullVersion))
print("Set %s to '%s'" %(strVersionFile, fullVersion))

###############################################################################
###  Prepare other files and construct compile arguments
###############################################################################

# get the long description from the README file.
# CAUTION: Python2/3 utf encoding shit calls needs some adjustments
fileName = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md')
f = (open(fileName, 'r') if sys.version_info < (3, 0)
     else open(fileName, 'r', encoding='utf-8'))
longDescription = f.read()
f.close()

# define different compiler arguments for each platform
strPlatform = platform.system()
compilerArguments = []
linkerArguments = []
if strPlatform == 'Windows':
    # Microsoft Visual C++ Compiler 9.0
    compilerArguments += ['/O2', '/fp:precise']
elif strPlatform == 'Linux':
    # assuming Linux and gcc
    compilerArguments += ['-O3', '-march=native', '-ffast-math']
elif strPlatform == 'Darwin':
    # assuming Linux and gcc
    compilerArguments += ['-O3', '-march=native', '-ffast-math']
else:
    WARNING("Your platform is currently not supported by %s: %s" % (
        packageName, strPlatform))

# define default cython directives, these may get extended along the script
cythonDirectives = {}
defineMacros = []
CMD_COVERAGE = '--enable-cython-tracing'
if CMD_COVERAGE in sys.argv:
    sys.argv.remove(CMD_COVERAGE)
    cythonDirectives['linetrace'] = True
    cythonDirectives['binding'] = True
    defineMacros += [('CYTHON_TRACE_NOGIL', '1'),
                     ('CYTHON_TRACE', '1')]
    print("Enabling cython line tracing allowing code coverage analysis")

print("Building %s v%s for %s." % (packageName, packageVersion, strPlatform))


###############################################################################
###  Enable flexible dependency handling by installing missing base components
###############################################################################
# override list type to allow lazy cythonization: cythonize and compile only
# after install_requires installed cython
class lazyCythonize(list):
    def __init__(self, callback):
        self._list, self.callback = None, callback

    def c_list(self):
        if self._list is None:
            self._list = self.callback()

        return self._list

    def __iter__(self):
        for e in self.c_list():
            yield e

    def __getitem__(self, ii):
        return self.c_list()[ii]

    def __len__(self):
        return len(self.c_list())


def extensions():
    try:
        from Cython.Build import cythonize
    except ImportError:
        def cythonize(*args, **kwargs):
            print("Hint: Wrapping import of cythonize in extensions()")
            from Cython.Build import cythonize
            return cythonize(*args, **kwargs)

    try:
        import numpy
        lstIncludes = [numpy.get_include()]
    except ImportError:
        lstIncludes = []

    extensionArguments = {
        'include_dirs':
        lstIncludes + ['fastmat/core', 'fastmat/inspect', 'util'],
        'extra_compile_args': compilerArguments,
        'extra_link_args': linkerArguments,
        'define_macros': defineMacros
    }

    return cythonize(
        [Extension("*", ["fastmat/*.pyx"], **extensionArguments),
         Extension("*", ["fastmat/algs/*.pyx"], **extensionArguments),
         Extension("*", ["fastmat/core/*.pyx"], **extensionArguments)],
        compiler_directives=cythonDirectives)


# determine requirements for install and setup
def checkRequirement(lstRequirements, importName, requirementName):
    '''
    Don't add packages unconditionally as this involves the risk of updating an
    already installed package. Sometimes this may break during install or mix
    up dependencies after install. Consider an update only if the requested
    package is not installed at all or if we are building an installation wheel.
    '''
    try:
        __import__(importName)
    except ImportError:
        lstRequirements.append(requirementName)
    else:
        if 'bdist_wheel' in sys.argv[1:]:
            lstRequirements.append(requirementName)


setupRequires = []
installRequires = []
checkRequirement(setupRequires, 'setuptools', 'setuptools>=18.0')
checkRequirement(setupRequires, 'Cython', 'cython>=0.19')
checkRequirement(setupRequires, 'numpy', 'numpy')
checkRequirement(installRequires, 'six', 'six')
checkRequirement(installRequires, 'scipy', 'scipy')

print("Requirements for setup: %s" %(setupRequires))
print("Requirements for install: %s" %(installRequires))

###############################################################################
### The actual setup
###############################################################################
setup(
    name=packageName,
    version=packageVersion,
    description='fast linear transforms in Python',
    long_description=longDescription,
    author='Christoph Wagner, Sebastian Semper, EMS group TU Ilmenau',
    author_email='christoph.wagner@tu-ilmenau.de',
    url='https://ems-tu-ilmenau.github.io/fastmat/',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: IPython',
        'Framework :: Jupyter',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries'
    ],
    keywords='linear transforms efficient algorithms mathematics',
    setup_requires=setupRequires,
    install_requires=installRequires,
    packages=[
        'fastmat',
        'fastmat/algs',
        'fastmat/core',
        'fastmat/inspect'
    ],
    ext_modules=lazyCythonize(extensions)
)



# -*- coding: utf-8 -*-
'''
  fastmat/Parametric.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for Parametric class (structural description).


  Author      : wcw
  Introduced  : 2016-11-09
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *

ctypedef object(*function)(object, object)

################################################################################
################################################## class Hadamard
cdef class Parametric(Matrix):

    ############################################## class variables
    cdef public np.ndarray  _vecX                # support vector of X-dimension
    cdef public np.ndarray  _vecY                # support vector of Y-dimension
    cdef public object      _fun                 # Parameterizing function

    cdef object      _funDtype                   # element function data type
    cdef bint        _rangeAccess                # if True, _fun will be called
    #                                            # for every element separately
    #                                            # (no range access calls)

    ############################################## class methods
    cdef void _core(self, np.ndarray, np.ndarray, ftype, ftype, ftype, bint)

    cpdef _forwardC(self, np.ndarray, np.ndarray, ftype, ftype)
    cpdef _backwardC(self, np.ndarray, np.ndarray, ftype, ftype)

    cpdef np.ndarray _reference(self)



# -*- coding: utf-8 -*-
'''
  fastmat/Sum.py
 -------------------------------------------------- part of the fastmat package

  Linear combination of matrices.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .Matrix cimport Matrix

################################################################################
################################################## class Sum
cdef class Sum(Matrix):

    ############################################## class methods
    def __init__(self, *matrices):
        '''
        Initialize Matrix instance with a list of other matrices to be summed.
        If another Sum is seen, add its content instead of adding the Sum.
        '''
        cpdef Matrix mat

        # Fold multiple levels of sums
        lstTerms = []

        def __addTerms(matrices):
            for mat in matrices:
                if not isinstance(mat, Matrix):
                    raise TypeError("Sum: Term is not a fastmat Matrix.")

                # flatten nested instances of Sum
                if isinstance(mat, Sum):
                    __addTerms(mat.content)
                else:
                    lstTerms.append(mat)

        __addTerms(matrices)
        self._content = tuple(lstTerms)

        # Have at least one term
        cdef int ii, cntTerms = len(self._content)
        if cntTerms < 1:
            raise ValueError("Sum: No terms given.")

        # check for matching transform dimensions
        cdef intsize numN = self._content[0].numN
        cdef intsize numM = self._content[0].numM
        for ii in range(1, cntTerms):
            mat = self._content[ii]
            if (mat.numN != numN) or (mat.numM != numM):
                raise ValueError("Sum: Term dimension mismatch: " + repr(mat))

        # determine data type of sum result
        dataType = np.int8
        for ii in range(0, cntTerms):
            dataType = np.promote_types(dataType, self._content[ii].dtype)

        # set properties of matrix
        self._initProperties(
            numN, numM, dataType,
            cythonCall=True,
            widenInputDatatype=True
        )

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef int cc, cnt = len(self._content)

        result = self._content[0]._getCol(idx).astype(self.dtype)
        for cc in range(1, cnt):
            result += self._content[cc]._getCol(idx)

        return result

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef int cc, cnt = len(self._content)

        result = self._content[0]._getRow(idx).astype(self.dtype)
        for cc in range(1, cnt):
            result += self._content[cc]._getRow(idx)

        return result

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        cdef int cc, cnt = len(self._content)

        result = self._content[0]._getItem(idxN, idxM).astype(self.dtype)
        for cc in range(1, cnt):
            result += self._content[cc]._getItem(idxN, idxM)

        return result

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = 0.
        cdef Matrix item

        for item in self._content:
            complexity += item.numN + item.numM

        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the forward transform of this matrix'''
        cdef int cc, cnt = len(self._content)

        arrRes[:] = self._content[0].forward(arrX)
        for cc in range(1, cnt):
            arrRes += self._content[cc].forward(arrX)

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the backward transform of this matrix'''
        cdef int cc, cnt = len(self._content)

        arrRes[:] = self._content[0].backward(arrX)
        for cc in range(1, cnt):
            arrRes += self._content[cc].backward(arrX)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes
        cdef int cc, cnt = len(self._content)

        arrRes = np.zeros((self.numN, self.numM), dtype=self.dtype)
        for cc in range(cnt):
            arrRes += self._content[cc].reference()

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 25,
                TEST.NUM_M      : TEST.Permutation([17, TEST.NUM_N]),
                'mType1'        : TEST.Permutation(TEST.ALLTYPES),
                'mType2'        : TEST.Permutation(TEST.FEWTYPES),
                'arrM1'         : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M)
                }),
                'arrM2'         : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M)
                }),
                'arrM3'         : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M)
                }),
                TEST.INITARGS   : [lambda param: Matrix(param['arrM1']()),
                                   lambda param: Matrix(param['arrM2']()),
                                   lambda param: Matrix(param['arrM3']())],
                TEST.OBJECT     : Sum,
                TEST.NAMINGARGS : dynFormat("%s+%s+%s",
                                            'arrM1', 'arrM2', 'arrM3'),
                TEST.TOL_POWER  : 3.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        from .Eye import Eye
        from .Fourier import Fourier
        from .Circulant import Circulant
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Sum(
                    Fourier(c), Eye(c), Circulant(np.random.randn(c)))),
                BENCH.FUNC_SIZE : (lambda c: c)
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Sum(*([Eye(2 ** c)] * c))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Sum of Matrices (\texttt{fastmat.Sum})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For matrices $\bm A_k \in \C^{n \times m}$ with $k = 1,\dots,N$ we define a new
mapping $\bm M$ as the sum \[\bm M = \Sum{k = 1}{N}{\bm A_k},\] which then also
is a mapping in $\C^{n \times m}$.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the components',
                            'A = fm.Circulant(x_A)',
                            'B = fm.Circulant(x_B)',
                            'C = fm.Fourier(n)',
                            'D = fm.Diag(x_D)',
                            '',
                            '# construct the sum of transformations',
                            'M = fm.Sum(A, B, C, D)',
                            caption=r"""
Assume we have two circulant matrices $\bm A$ and $\bm B$, an $N$-dimensional
Fourier matrix $\bm C$ and a diagonal matrix $\bm D$. Then we define
\[\bm M = \bm A + \bm B + \bm C + \bm D.\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
The \emph{forward} and \emph{solve} benchmarks were performed on a matrix
$\bm L = \bm \Fs_{k} + \bm I_{k} + \bm C_{k}$
whereas the \emph{overhead} benchmark was performed on a matrix
$\bm L = \sum\limits_{i = 1}^{k} \bm I_{2^k}$; so $n = 2^k$ for $k \in \N$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Outer.pyx
 -------------------------------------------------- part of the fastmat package

  Outer product (rank one) matrix.


  Author      : sempersn
  Introduced  : 2017-31-01
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.cmath cimport _conjugate, _multiply

################################################################################
################################################## class Outer
cdef class Outer(Matrix):

    ############################################## class properties
    # vecV - Property (read-only)
    # Return the matrix-defining vector of vertical defining entries.
    property vecV:
        def __get__(self):
            return self._vecV

    # vecH - Property (read-only)
    # Return the matrix-defining vector of horizontal defining entries.
    property vecH:
        def __get__(self):
            return self._vecH

    ############################################## class methods
    def __init__(self, vecV, vecH):

        # check dimensions
        vecV = np.atleast_1d(np.squeeze(vecV))
        vecH = np.atleast_1d(np.squeeze(vecH))

        if vecV.ndim != 1 or vecH.ndim != 1:
            raise ValueError("Outer parameters must be one-dimensional.")

        # height and width of matrix is defined by length of input vectors
        cdef intsize numN = len(vecV)
        cdef intsize numM = len(vecH)

        # determine joint data type of operation
        datatype = np.promote_types(vecV.dtype, vecH.dtype)

        # store H/V entry vectors as copies of vecH/vecV
        self._vecH = vecH.astype(datatype, copy=True).reshape((1, numM))
        self._vecV = vecV.astype(datatype, copy=True).reshape((numN, 1))

        # store ravelled pointers of vecH/vecV
        self._vecHRav = self._vecH.ravel()
        self._vecVRav = self._vecV.ravel()

        # hermitian transpose of vecH/vecV for backward
        self._vecHConj = _conjugate(self._vecH).reshape((numM, 1))
        self._vecVHerm = _conjugate(self._vecV).reshape((1, numN))

        # set properties of matrix
        self._initProperties(numN, numM, datatype)

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        return self._vecHRav[idx] * self._vecVRav

    cpdef np.ndarray _getRow(self, intsize idx):
        return self._vecVRav[idx] * self._vecHRav

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._vecVRav[idxN] * self._vecHRav[idxM]

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = 2 * (self.numN + self.numM)
        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix.'''

        return self._vecV.dot(self._vecH.dot(arrX))

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix.'''

        return self._vecHConj.dot(self._vecVHerm.dot(arrX))

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        dtype = np.promote_types(self.dtype, np.float64)
        return self._vecV.dot(self._vecH.astype(dtype))

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 4,
                TEST.NUM_M      : TEST.Permutation([6, TEST.NUM_N]),
                'mTypeH'        : TEST.Permutation(TEST.FEWTYPES),
                'mTypeV'        : TEST.Permutation(TEST.ALLTYPES),
                'vecH'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeH',
                    TEST.SHAPE  : (TEST.NUM_N, 1)
                }),
                'vecV'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeV',
                    TEST.SHAPE  : (TEST.NUM_M, 1)
                }),
                TEST.INITARGS   : (lambda param : [param['vecH'](),
                                                   param['vecV']()]),
                TEST.OBJECT     : Outer,
                TEST.NAMINGARGS : dynFormat("%so%s", 'vecH', 'vecV'),
                TEST.TOL_MINEPS : (lambda param:
                                   max(_getTypeEps(param['mTypeH']),
                                       _getTypeEps(param['mTypeV']))),
                TEST.TOL_POWER  : 4.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {
                # ignore int8 datatype as there will be overflows
                TEST.IGNORE     : TEST.IgnoreFunc(lambda param: (
                    param['mTypeH'] == param['mTypeV'] ==
                    param[TEST.DATATYPE] == np.int8))
            }
        }

    def _getBenchmark(self):
        from .inspect import BENCH, arrTestDist
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Outer(
                    arrTestDist((c, ), dtype=np.float),
                    arrTestDist((c, ), dtype=np.float)))
            },
            BENCH.FORWARD: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Outer(
                    arrTestDist((2 ** c, ), dtype=np.float),
                    arrTestDist((2 ** c, ), dtype=np.float)))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, datatype: Outer(
                    arrTestDist((2 ** c, ), dtype=datatype),
                    arrTestDist((2 ** c, ), dtype=datatype)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Outer Product Matrix (\texttt{fastmat.Outer})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
The outer product is a special case of the Kronecker product of one-dimensional
vectors. For given $\bm a \in \C^n$ and $\bm b \in \C^m$ it is defined as
    \[\bm x \mapsto \bm a \cdot \bm b^\trans \cdot \bm x.\]
It is clear, that this matrix has at most rank $1$ and as such has a fast
transformation.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# define parameter',
                            'n, m = 4, 5',
                            'v = np.arange(n)',
                            'h = np.arange(m)',
                            '',
                            '# construct the matrix',
                            'M = fm.Outer(v, h)',
                            caption=r"""
This yields
\[\bm v = (0,1,2,3,4)^T\]
\[\bm h = (0,1,2,3,4,5)^T\]
\[\bm M = \left(\begin{array}{ccccc}
    0 & 0 & 0 & 0 & 0  \\
    0 & 1 & 2 & 3 & 4  \\
    0 & 2 & 4 & 6 & 8  \\
    0 & 3 & 6 & 9 & 12
\end{array}\right)\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix $\bm M \in \R^{n \times n}$ with
defining vectors' entries drawn from a uniform distribution on $[2,3]$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/BlockDiag.py
 -------------------------------------------------- part of the fastmat package

  Block diagonal matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - BlockDiag should simply skip all Zero Matrices (flag them as "None")?
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *

################################################################################
################################################## class BlockDiag
cdef class BlockDiag(Matrix):

    ############################################## class methods
    def __init__(self, *matrices):
        '''Initialize Matrix instance with a list of child matrices'''
        cdef intsize numN = 0, numM = 0
        cdef Matrix term

        self._content = matrices

        # determine total size and data type of matrix
        dataType = np.int8
        for term in self._content:
            if not isinstance(term, Matrix):
                raise ValueError(
                    "Only fastmat matrices supported, %s given." %(str(term)))

            numN += term.numN
            numM += term.numM
            dataType = np.promote_types(dataType, term.dtype)

        # set properties of matrix
        self._initProperties(
            numN, numM, dataType,
            cythonCall=True,
            widenInputDatatype=True
        )

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexityFwd = self.numN
        cdef float complexityBwd = self.numM
        cdef Matrix item
        for item in self:
            complexityFwd += item.numN + item.numM
            complexityBwd += item.numM + item.numN

        return (complexityFwd, complexityBwd)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the forward transform of this matrix'''
        cdef Matrix term
        cdef intsize idxN = 0, idxM = 0, ii, cnt = len(self._content)

        for ii in range(0, cnt):
            term = self._content[ii]
            arrRes[idxN:(idxN + term.numN), :] \
                = term.forward(arrX[idxM:(idxM + term.numM)])

            idxN += term.numN
            idxM += term.numM

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''Calculate the backward transform of this matrix'''
        cdef Matrix term
        cdef intsize idxN = 0, idxM = 0, cnt = len(self._content)

        for ii in range(0, cnt):
            term = self._content[ii]

            arrRes[idxM:(idxM + term.numM), :] \
                = term.backward(arrX[idxN:(idxN + term.numN)])

            idxN += term.numN
            idxM += term.numM

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes
        cdef Matrix term
        cdef intsize idxN = 0, idxM = 0

        arrRes = np.zeros((self.numN, self.numM), dtype=self.dtype)

        for term in self._content:
            arrRes[idxN:(idxN + term.numN), :][:, idxM:(idxM + term.numM)] = \
                term._getArray()

            idxN += term.numN
            idxM += term.numM

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                'size'          : 5,
                TEST.NUM_N      : (lambda param: param['size'] * 3),
                TEST.NUM_M      : TEST.NUM_N,
                'mType1'        : TEST.Permutation(TEST.ALLTYPES),
                'mType2'        : TEST.Permutation(TEST.ALLTYPES),
                'arr1'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : ('size', 'size')
                }),
                'arr2'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType2',
                    TEST.SHAPE  : ('size', 'size')
                }),
                'arr3'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mType1',
                    TEST.SHAPE  : ('size', 'size')
                }),
                TEST.INITARGS   : (lambda param: [Matrix(param['arr1']()),
                                                  -2. * Matrix(param['arr2']()),
                                                  2. * Matrix(param['arr3']())
                                                  ]),
                TEST.OBJECT: BlockDiag,
                'strType1'      : (lambda param:
                                   TEST.TYPENAME[param['mType1']]),
                'strType2'      : (lambda param:
                                   TEST.TYPENAME[param['mType2']]),
                TEST.NAMINGARGS : dynFormat("%s,%s,%s:(%dx%d) each",
                                            'strType1', 'strType2', 'strType1',
                                            'size', 'size'),
                TEST.TOL_POWER  : 3.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH, arrTestDist
        from .Circulant import Circulant
        from .Diag import Diag
        from .Eye import Eye
        from .Fourier import Fourier
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: BlockDiag(
                    Circulant(np.random.randn(c)),
                    Circulant(np.random.randn(c)),
                    Fourier(c), Diag(np.random.randn(c))
                )),
                BENCH.FUNC_SIZE : (lambda c: 4 * c)
            },
            BENCH.FORWARD: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: BlockDiag(*([Eye(2 ** c)] * 16))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c * 16)
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Block Diagonal Matrix (\texttt{fastmat.BlockDiag})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
\[\bm M = \mathrm{diag}\left\{\left( \bm A_{i}\right)_{i}\right\},\]
where the $\bm A_{i}$ can be fast transforms of \emph{any} type.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the blocks',
                            'A = fm.Circulant(x_A)',
                            'B = fm.Circulant(x_B)',
                            'C = fm.Fourier(n)',
                            'D = fm.Diag(x_D)',
                            '',
                            '# define the block',
                            '# diagonal matrix',
                            'M = fm.BlockDiag(A, B, C, D)',
                            caption=r"""
Assume we have two circulant matrices $\bm A$ and $\bm B$, an $N$-dimensional
Fourier matrix $\bm C$ and a diagonal matrix $\bm D$. Then we define
\[\bm M = \left(\begin{array}{cccc}
    \bm A & & & \\
    & \bm B & & \\
    & & \bm C & \\
    & & & \bm D
\end{array}\right).\]"""),
                r"""
Meta types can also be nested, so that a block diagonal matrix can contain
products of block matrices as its entries. Note that the efficiency of the fast
transforms decreases the more building blocks they have.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the blocks',
                            'A = fm.Circulant(x_A)',
                            'B = fm.Circulant(x_B)',
                            'F = fm.Fourier(n)',
                            'D = fm.Diag(x_D)',
                            '',
                            '# define a product',
                            'P = fm.Product(A.H, B)',
                            '',
                            '# define the block',
                            '# diagonal matrix',
                            'M = fm.BlockDiag(P, F, D)',
                            caption=r"""
Assume we have a product $\bm P$ of two matrices $\bm A^\herm$ and $\bm B$, an
$N$-dimensional Fourier matrix $\bm{\mathcal{F}}$ and a diagonal matrix
$\bm D$. Then we define
\[\bm M = \left(\begin{array}{cccc}
    \bm A^\herm \cdot \bm B &                  &        \\
                            & \bm{\mathcal{F}} &        \\
                            &                  & \bm D
\end{array}\right).\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks',
                DOC.PLOTFORWARD(doc=r'(Matrix as in snippet)'),
                DOC.PLOTFORWARDMEMORY(doc=r'(same as in snippet)'),
                DOC.PLOTOVERHEAD(doc=r"""
$\bm B = \begin{pmatrix}
    \bm I_{2^k} & \dots     & 0             \\
                & \ddots    &               \\
    0           & \dots     & \bm I_{2^k}
\end{pmatrix}$"""),
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/core/types.pxd
 -------------------------------------------------- part of the fastmat package

  fastmat type handling support.

  Author      : wcw
  Introduced  : 2016-11-16
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
from libc.stdlib cimport malloc

import numpy as np
cimport numpy as np

# initialize numpy C-API interface for cython extension-type classes.
# Theis call is required once for every module that uses PyArray_ calls
# WARNING: DO NOT REMOVE THIS LINE OR SEGFAULTS ARE ABOUT TO HAPPEN!
np.import_array()

################################################## type handling

cdef inline void _getDtypeInfo(np.dtype dtype, INFO_TYPE_s *info):
    '''Fill in type descriptor (INTO_TYPE_s) for given a numpy.dtype.'''
    info[0].typeNum     = dtype.type_num
    info[0].fusedType   = typeSelection[info[0].typeNum]
    info[0].dsize       = dtype.itemsize
    info[0].isComplex   = np.PyDataType_ISCOMPLEX(dtype)
    info[0].isInt       = np.PyDataType_ISINTEGER(dtype)
    info[0].isNumber    = np.PyDataType_ISNUMBER(dtype)

    info[0].eps         = 0
    info[0].min         = 0
    info[0].max         = 0
    if dtype != np.bool:
        if info[0].isNumber and not info[0].isInt:
            # floating-point types
            dtypeInfo       = np.finfo(dtype)
            info[0].eps     = dtypeInfo.eps
            info[0].min     = dtypeInfo.min
            info[0].max     = dtypeInfo.max

        if info[0].isInt:
            # integer-based types
            dtypeInfo       = np.iinfo(dtype)
            info[0].min     = dtypeInfo.min
            info[0].max     = dtypeInfo.max

    # by default, no promotion array will be present
    # also, don't allocate memory for it
    info[0].promote     = NULL


cdef INFO_TYPE_s *_getTypeInfo(object dtype):
    '''
    Return a pointer to a type descriptor (INFO_TYPE_s) for a given type
    object. Takes any valid type descriptor as input (np.dtype, int, type).
    This function fetches actual data from the pre-generated 'type_info' array.
    '''
    cdef np.dtype nptype
    if type(dtype) == type:
        # if dtype is a python type object, convert it to a numpy type container
        nptype = np.dtype(dtype)
    elif type(dtype) == int:
        # then check for common types: np.dtype and int (numpy typenum)
        nptype = np.PyArray_DescrFromType(dtype)
    elif not isinstance(dtype, np.dtype):
        # throw an error if itself not a dtype
        raise TypeError("Invalid type information %s" % (str(dtype)))
    else:
        nptype = dtype

    cdef ftype fusedTypeNum = typeSelection[nptype.type_num]
    if fusedTypeNum == TYPE_INVALID:
        raise TypeError("Not a fastmat fused type: %s" %(str(dtype)))

    return &(typeInfo[fusedTypeNum])


cdef ftype _getFusedType(INFO_TYPE_s * dtype):
    '''Return most suited fastmat datatype id for given type descriptor.'''
    if dtype[0].isNumber:
        if dtype[0].isInt:
            if dtype[0].dsize <= 1:
                return TYPE_INT8
            elif dtype[0].dsize <= 4:
                return TYPE_INT32
            elif dtype[0].dsize <= 8:
                return TYPE_INT64
            else:
                return TYPE_FLOAT64
        elif dtype[0].isComplex:
            return TYPE_COMPLEX64 if (dtype[0].dsize <= 8) else TYPE_COMPLEX128
        else:
            if dtype[0].dsize == 1:
                # bool ends here as it is a number, but neither int nor complex
                return TYPE_INT8
            else:
                return TYPE_FLOAT32 if (dtype[0].dsize <= 4) else TYPE_FLOAT64

    return TYPE_INVALID


cdef nptype _getNpType(np.ndarray arr):
    '''Return numpy type number of array'''
    return np.PyArray_TYPE(arr)


cdef ftype _getFType(np.ndarray arr):
    '''Return fused type number of ndarray arr.'''
    return typeSelection[np.PyArray_TYPE(arr)]


cpdef np.float64_t _getTypeEps(object dtype):
    '''Return eps for a given (and known) type.'''
    cdef INFO_TYPE_s *info = _getTypeInfo(dtype)
    return info[0].eps


cpdef np.float64_t _getTypeMin(object dtype):
    '''Return the minimum representable value for a given (and known) type.'''
    cdef INFO_TYPE_s *info = _getTypeInfo(dtype)
    return info[0].min


cpdef np.float64_t _getTypeMax(object dtype):
    '''Return the maximum representable value for a given (and known) type.'''
    cdef INFO_TYPE_s *info = _getTypeInfo(dtype)
    return info[0].max


################################################## type Promotion stuff

cdef ftype _promoteTypeNums(ftype type1, ftype type2):
    if (type1 < 0) or (type2 >= NUM_TYPES) or \
       (type2 < 0) or (type2 >= NUM_TYPES):
        raise ValueError("Invalid type numbers for promotion")

    return typeInfo[type1].promote[type2]

cpdef object safeTypeExpansion(object dtype):
    return (np.float32 if dtype == np.int8
            else (np.float64 if (dtype == np.int32) or (dtype == np.int64)
                  else dtype))


### put some type promotion code here! Initialize type table list, specifying
### type conversion rules for this very machine. Define a number of basic types
### which are fused to declare different implementation specializations. Then,
### handle promotion by simple table-lookups in simple steps:
###   (1)  break down the multitude of different types, classify unsupported
###        unsupported ones
###   (2)  generate a type promotion matrix for the remaining types
###   (3)  translate type back for array generation
###
### After a type is correctly promoted and its corresponding result array cre-
### ated, also extract memoryviews with these types for the call specialization
### following.
###
### np.NPY_NTYPES is equal to the total number of numpy types available
### NUM_TYPES is equal to the number of fused types supported (incl. ~INVALID)
### typeInfo contains information about each fused type used in fastmat
### typeSelection can be used to fit numpy data types to fused types

# definine association between numpy data types to fused data types
# CAUTION: must be initialized before typeInfo as the fusedType ID will be taken
#          from typeSelection[]
cdef INFO_TYPE_s ttDescr
cdef np.dtype numpyDtype
cdef int tt, ii

# determine suitable type number for numpy types
for tt in range(<int> np.NPY_NTYPES):
    # fusedType field will not be correct as typeSelection is not yet filled.
    _getDtypeInfo(np.PyArray_DescrFromType(tt), &ttDescr)
    typeSelection[tt] = _getFusedType(&ttDescr)

# set fused type definitions for specific numpy data types
for tt, dtype in {
    TYPE_INT8: np.int8,
    TYPE_INT32: np.int32,
    TYPE_INT64: np.int64,
    TYPE_FLOAT32: np.float32,
    TYPE_FLOAT64: np.float64,
    TYPE_COMPLEX64: np.complex64,
    TYPE_COMPLEX128: np.complex128,
    TYPE_INVALID: np.void
}.items():
    _getDtypeInfo(np.PyArray_DescrFromTypeObject(dtype), &(typeInfo[tt]))

# initialize type promotion stuff:
for tt in range(NUM_TYPES):
    typeInfo[tt].promote = <ftype *> malloc(sizeof(int) * NUM_TYPES)
    for ii in range(NUM_TYPES):
        t1 = np.PyArray_TypeObjectFromType(typeInfo[tt].typeNum)
        t2 = np.PyArray_TypeObjectFromType(typeInfo[ii].typeNum)
        if tt < <int> TYPE_INVALID and ii < <int> TYPE_INVALID:
            numpyDtype = np.promote_types(t1, t2)
            typeInfo[tt].promote[ii] = typeSelection[numpyDtype.type_num]
        else:
            numpyDtype = np.dtype(np.void)
            typeInfo[tt].promote[ii] = TYPE_INVALID

# debug output: show internal data structures


def _typeSelection():
    '''Print internals to assist debugging.'''
    print("typeSelection =")
    lst = [[] for ii in range(NUM_TYPES)]
    for ii in range(<int> np.NPY_NTYPES):
        lst[typeSelection[ii]].append(
            "%d:%s" % (ii, str(np.PyArray_DescrFromType(ii))))

    for ii in range(NUM_TYPES):
        print("\t%d:%s" % (ii, lst[ii]))


def _typeInfo():
    print("typeInfo =")
    for tt in range(NUM_TYPES):
        lst = [typeInfo[tt].promote[ii] for ii in range(NUM_TYPES)]
        ii = typeInfo[tt].typeNum
        print("\t%d:%s" % (tt, {
            'typeNum'   : "%d:%s" % (ii, str(np.PyArray_DescrFromType(ii))),
            'fusedType' : typeInfo[tt].fusedType,
            'dsize'     : typeInfo[tt].dsize,
            'isComplex' : typeInfo[tt].isComplex,
            'isInt'     : typeInfo[tt].isInt,
            'isNumber'  : typeInfo[tt].isNumber,
            'promote'   : lst,
            'eps'       : typeInfo[tt].eps,
            'min'       : typeInfo[tt].min,
            'max'       : typeInfo[tt].max
        }))



# -*- coding: utf-8 -*-
'''
  demo/lowRankApprox.py
 -------------------------------------------------- part of the fastmat demos

  Demonstration how to do low rank approximations of dense matrices
  that have a rapidly decaying spectrum.

  Author      : sempersn
  Introduced  : 2016-09-28
 ------------------------------------------------------------------------------
  PARAMETERS:
    numMatSize      - size of the system matrix
    numApproxQual   - approximation quality in [0,1]

 ------------------------------------------------------------------------------
   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

import sys
import time

# import numpy functionality
import numpy.random as npr
import numpy.linalg as npl
import numpy as np

################################################## import modules
try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat

################################################################################
#                          PARAMETER SECTION
################################################################################
# please edit these parameters as you please, but be cautious with the problem
# size, because the traditional method might fail or take a long time
################################################################################
numMatSize = 2 ** 9
numApproxQual = 0.3

################################################################################
#                          CALCULATION SECTION
################################################################################

print("fastmat demo: Approximation of an almost low rank matrix with LowRank()")
print("-----------------------------------------------------------------------")

# calc the number of dimensions to approximate
numSize = int(numApproxQual * numMatSize)

# draw a dense random and normalized array
print(" * Generate the Matrix")
arrFull = npr.randn(numMatSize, numMatSize) / np.sqrt(numMatSize)

# get the SVD
print(" * Calc the SVD")
arrU, vecSigma, arrV = npl.svd(arrFull)

print(" * Truncate the singular values")
vecT = np.linspace(0, 3, numMatSize - numSize - 1)
vecSigma[numSize + 1:] = vecSigma[numSize + 1:] * 0.1 * np.exp(-vecT)

print(" * Rebuild the matrix with truncated singular values")
arrFull = arrU.dot( np.diag(vecSigma).dot(arrV.T) )

matFull = fastmat.Matrix(arrFull)
matApprox = fastmat.LowRank(
    vecSigma[:numSize],arrU[:,:numSize],arrV[:,:numSize]
)

print(" * Generate the linear system")
vecX = npr.randn(numMatSize)
vecB = matFull * vecX

s = time.time()
y1 = matFull * vecX
timeDenseFwd = time.time() - s

s = time.time()
y2 = matApprox * vecX
timeApproxFwd = time.time() - s

s = time.time()
x1 = fastmat.algs.CG(matFull,vecB)
timeDenseSolve = time.time() - s

s = time.time()
x2 = fastmat.algs.CG(matApprox,vecB)
timeApproxSolve = time.time() - s

numApproxErr1 = npl.norm(x1 - x2) / npl.norm(x1)
numApproxErr2 = npl.norm(matFull * x1 - matFull * x2) / npl.norm(vecB)

################################################################################
#                               OUTPUT SECTION
################################################################################
print("\nResults:")
print("   Dense Multiplication             %14.3f ms" %(1e3 * timeDenseFwd))
print("   Approximated Multiplication      %14.3f ms" %(1e3 * timeApproxFwd))
print("   Solve Dense Linear System        %14.3f ms" %(1e3 * timeDenseSolve))
print("   Solve Approximated Linear System %14.3f ms" %(1e3 * timeApproxSolve))
print("   Relative Error in Coefficient Domain %10.3f" %(numApproxErr1))
print("   Relative Error in Signal Domain      %10.3f" %(numApproxErr2))



# -*- coding: utf-8 -*-
'''
  fastmat/core/resource.py
 -------------------------------------------------- part of the fastmat package

  Routines for resource handling and estimation of consumption.


  Author      : wcw
  Introduced  : 2017-01-05
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import sys
from itertools import chain
from collections import deque
import numpy as np
from scipy.sparse import spmatrix
from ..Matrix import Matrix

################################################## getMemoryFootprint()


def getMemoryFootprint(obj, **options):
    '''
    Return the approximate memory footprint of an object with all of its
    contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
            OtherContainerClass: OtherContainerClass.get_elements}
    '''
    # extract options
    handlers = options.get('handlers', {})
    verbose = options.get('verbose', False)

    # keep track of object ids already seen
    seen = set()

    # estimate size of objects without __sizeof__()
    default_size = sys.getsizeof(0)

    # setup handlers for various data types
    def dict_handler(d):
        return chain.from_iterable(d.items())

    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)

    # walker for getting size of object considering corner and special cases
    def sizeof(obj):
        # skip objects already seen once
        if id(obj) in seen:
            return 0

        seen.add(id(obj))

        s = sys.getsizeof(obj, default_size)

        # check fastmat matrices:
        # only cdefs with explicit `public` tag will be seen here !
        if isinstance(obj, Matrix):
            for key in dir(obj):
                if key[0] == '_' and key[1] != '_':
                    item = getattr(obj, key)
                    if not callable(item) and (item is not None):
                        s += sizeof(item)

        # check for ndarrays (have special properties holding data)
        if isinstance(obj, np.ndarray):
            if obj.base is not None:
                # add memory size of base (if not yet added)
                # some numpy versions don't report properly to getsizeof()
                added = sizeof(obj.base)
                return s + added if added > s else s

            # some numpy versions don't report properly to getsizeof()
            added = obj.nbytes
            s += added if added > s else 0

        # check for sparse arrays (have special properties holding data)
        if isinstance(obj, spmatrix):
            s += sizeof(obj.__dict__)

        if verbose:
            print("==%d, [%s@%x]" %(s, type(obj), id(obj)))

        # check for known container types
        for typ, handler in all_handlers.items():
            if isinstance(obj, typ):
                size = sum(map(sizeof, handler(obj)))
                s += size
                break

        return s

    return sizeof(obj)



# -*- coding: utf-8 -*-
'''
  fastmat/Permutation.pyx
 -------------------------------------------------- part of the fastmat package

  Permutation matrix.

  Author      : sempersn
  Introduced  : 2017-06-30
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *

################################################################################
################################################## class Permutation
cdef class Permutation(Matrix):

    property sigma:
        def __get__(self):
            return self._sigma

    ############################################## class methods
    def __init__(self, sigma):
        '''
        Initialize Matrix instance with its dimensions.

        Generated a [numN x numM] matrix of small integers
        '''
        numN = sigma.shape[0]

        if not np.allclose(np.sort(sigma), np.arange(numN)):
            raise ValueError("Not a permutation.")

        self._sigma = sigma
        self._tau = np.argsort(sigma)

        self._initProperties(numN, numN, np.int8)

    ############################################## class property override
    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return 1 if (self._sigma[idxN] == idxM) else 0

    cpdef np.ndarray _getArray(self):
        return np.eye(self.numN, dtype=self.dtype)[self.sigma, :]

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 1.

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 1.

    cpdef Matrix _getNormalized(self):
        return self

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (self.numN, self.numM)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        return arrX[self._sigma, :]

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        return arrX[self._tau, :]

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        return np.eye(self.numN, dtype=self.dtype)[self._sigma, :]

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 35,
                TEST.NUM_M      : TEST.NUM_N,
                TEST.OBJECT     : Permutation,
                TEST.INITARGS   : (lambda param :
                                   [np.random.permutation(param[TEST.NUM_N])])
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Permutation(np.random.permutation(c)))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Permutation(np.random.permutation(2 ** c)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r"Permutation Matrix (\texttt{fastmat.Permutation})",
            DOC.SUBSUBSECTION(
                "Definition and Interface", r"""
For a given permutation $\sigma \in S_n$ and a vector $\bm x \in \C^n$ we map
\[\bm x \mapsto \left(x_{\sigma(i)}\right)_{i = 1}^n.\]""",
                DOC.SNIPPET('# import the package',
                            'import fastmat',
                            '',
                            '# set the permutation',
                            'sigma = np.array([3,1,2,0])',
                            '',
                            '# construct the identity',
                            'P = fastmat.Permutation(sigma)',
                            caption=r"""
\[\bm J = \left(\begin{array}{cccc}
    0 & 0 & 0 & 1 \\
    0 & 1 & 0 & 0 \\
    0 & 0 & 1 & 0 \\
    1 & 0 & 0 & 0
\end{array}\right)\]""")
            ),
            DOC.SUBSUBSECTION(
                "Performance Benchmarks", r"""
All benchmarks were performed on a Matrix $\bm P_n$ with random permutation.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/inspect/benchmark.py
 -------------------------------------------------- part of the fastmat package

  Implementations for class benchmarking (quality management).


  Author      : wcw
  Introduced  : 2017-07-19
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import os
import numpy as np
from timeit import timeit
import time

from .common import *
from ..core.cmath import profileCall
from ..core.resource import getMemoryFootprint
from ..Matrix import Matrix
from ..algs.CG import CG
from ..version import __version__


################################################## BENCH CONSTANT-def class
class BENCH(NAME):
    FORWARD         = 'forward'
    SOLVE           = 'solve'
    OVERHEAD        = 'overhead'
    DTYPES          = 'dtypes'
    PERFORMANCE     = 'performance'

    FUNC_INIT       = 'funcInit'
    FUNC_GEN        = 'funcConstr'
    FUNC_SIZE       = 'funcSize'
    FUNC_STEP       = 'funcStep'

    STEP_START_K    = 'startK'
    STEP_LOGSTEPS   = 'logSteps'
    STEP_MINITEMS   = 'minItems'

    LIMIT_MEMORY    = 'maxMem'
    LIMIT_INIT      = 'maxInit'
    LIMIT_ITER      = 'maxIter'
    LIMIT_SIZE      = 'maxSize'

    RESULT_MEMORY   = 'totalMem'
    RESULT_INIT     = 'initTime'
    RESULT_ITER     = 'iterTime'
    RESULT_K        = 'K'
    RESULT_SIZE     = 'numN'
    RESULT_OVH_NESTED_F = 'ovhFwd'
    RESULT_OVH_NESTED_B = 'ovhBwd'
    RESULT_EFF_NESTED_F = 'effFwd'
    RESULT_EFF_NESTED_B = 'effBwd'
    RESULT_COMPLEXITY_F = 'cpxFwd'
    RESULT_COMPLEXITY_B = 'cpxBwd'
    RESULT_ESTIMATE_FWD = 'estFwd'
    RESULT_ESTIMATE_BWD = 'estBwd'
    RESULT_ITEMS    = '#'

    MEAS_MINTIME    = 'measMinTime'
    MEAS_MINREPS    = 'measMinReps'
    MEAS_MINRUNS    = 'measMinRuns'
    MEAS_COMPACT    = 'compactStats'


################################################################################
##################################################  benchmark implementations

################################################## weightedPercentile()
def weightedPercentile(data, **options):
    """
    O(nlgn) implementation for weighted_percentile, with linear interpolation
    #between weights.

    date:       Aug 25 '16 @ 13:37
    user:       nayyrv [Aug 16 '15 @ 10:52]
    source:     www.stackoverflow.com/questions/
                    21844024/weighted-percentile-using-numpy
    ``  This is my function, it give identical behaviour to `np.percentile(
        np.repeat(data, weights), percentile)` With less memory overhead,
        np.percentile is an O(n) implementation so it's potentially faster for
        small weights. It has all the edge cases sorted out - it's an exace
        solution. The interpolation answers above assume linear, when it's a
        step for most of he case, except when the weight is 1.

        Say we have data [1,2,3] with weights [3, 11, 7] and I want the
        25% percentile. My ecdf is going to be [3, 10, 21] and I'm looking for
        the 5th value. The interpolation will see [3, 1] and [10, 2] as the
        matches and interpolate gving 1.28 despite being entirely in the
        2nd bin with a value of 2.
    ``

    """
    # extract options
    percentile = np.array(options.get('percentile', [75, 25])) / 100.
    weights = options.get('weights', None)

    if weights is None:
        weights=np.ones(data.shape)

    dataIndSort=np.argsort(data)
    dataSort=data[dataIndSort]
    weights_sort=weights[dataIndSort]
    ecdf=np.cumsum(weights_sort)
    percentilePos=percentile * (weights.sum() - 1) + 1

    # need the 1 offset at the end due to ecdf not starting at 0
    locations=np.searchsorted(ecdf, percentilePos)
    outPercentiles=np.zeros(percentilePos.shape)

    for i, empiricalLocation in enumerate(locations):
        # iterate across the requested outPercentiles
        if ecdf[empiricalLocation - 1] == np.floor(percentilePos[i]):
            # i.e. is the percentile in between two separate values
            uppWeight=percentilePos[i] - ecdf[empiricalLocation - 1]
            lowWeight=1 - uppWeight

            outPercentiles[i]=lowWeight * dataSort[empiricalLocation - 1] + \
                uppWeight * dataSort[empiricalLocation]
        else:
            # i.e. the percentile is entirely in one bin
            outPercentiles[i]=dataSort[empiricalLocation]

    return outPercentiles


################################################## timingStats()
def timingStats(
    stats,
    compactStats=False
):
    '''
    Return the runtime statistics of the given list of measurements.
    Each measurement is stored as such a tuple:
    (averageRuntime, totalRuntime, numberRepetitions)

    A dictionary with clearly named statistical properties is returned.
    '''

    # accumulate all individual stats
    times=np.array([stat['avg'] for stat in stats], dtype=np.double)
    weights=np.array([stat['cnt'] for stat in stats], dtype=np.double)

    # calculate som stats
    reps=np.sum(weights)
    avg=np.average(times, weights=weights)
    minimum=np.min(times)

    # prepare output
    result={'min': minimum}

    # if compactStats is True, skip further stats
    if compactStats:
        return result

    # calculate weighted variance
    var=sum(weights * ((times - avg) ** 2)) / (reps * (1 - 1 / len(stats)))

    # calculate quartiles
    percs=[25, 75]
    wp=weightedPercentile(times, percentile=percs, weights=weights)

    result.update({'var': var, 'cnt': reps, 'min': minimum})
    result.update({"p%d" %(percs[i]): wp[i] for i in range(len(percs))})

    return result


################################################## timeCall()
def timeCall(call, *args, **options):
    '''
    Return runtime statistics for execution of a parameterizable method call
    specified by call(*args).

    Behaviour of the measurements can be adjusted with these options:
      minMeasTime       - minimum total time the evaluation shall take
      minRepetitions    - minimum number of calls needed for evaluation
      minRuns           - minimum number of single measurements (of one or more
                          repetitions each) required
    '''
    stats=[]

    # begin with one repetition and approach accumulated runtime (step[1])
    num=1
    step={'avg': 0., 'time': 0., 'cnt': num}
    while step['time'] < (options[BENCH.MEAS_MINTIME] * num) / (num + 1) * 0.9:
        step=profileCall(num, call, *args)
        num=num * 2 if step['time'] < 1e-6 \
            else max(1, int(options[BENCH.MEAS_MINTIME] / step['avg']))

    # store as first run with [num] reps
    stats.append(step)

    # perform repeated measurement to reach at least MIN_ITERATIONS
    runs=max(int(np.ceil((options[BENCH.MEAS_MINREPS] - num) / num)),
             options[BENCH.MEAS_MINRUNS] - 1)
    if runs > 0:
        stats.extend([profileCall(num, call, *args) for ii in range(0, runs)])

    return timingStats(stats, options[BENCH.MEAS_COMPACT])


################################################## timeCalls()
def timeCalls(calls, **options):
    '''
    wrapper for running test runtime measurements for a number of calls

        target      - targeted resolution of measurement
        calls       - list of dics describing each call with these entries:
            > 'func'   (method pointer)
            > 'args'   (*list of arguments)
            > 'kwargs' (**dictionary of keyword-args)
        options     - dictionary with parameters for measurements

    A dictionary containing the measurement results of each call, preceded with
    its call tag is returned. All returned keys will be in `camelCase`.
    '''
    return {'%s%s' % (tag, key.title()): value
            for tag, call in calls.items()
            for key, value in (
                timeCall(call['func'], *(call['args']), **options)).items()}


################################################################################
##################################################  TESTS: specification

################################################## testInitSolve()
def testInitSolve(funcConstr, numSize, numN):
    instance=funcConstr(numSize)
    mem1=instance.nbytes

    func1=CG
    func2=np.linalg.solve

    args1=[instance, arrTestDist((numN, 1), instance.dtype)]
    args2=[instance._forwardReference(np.eye(numN)), args1[1]]

    if instance._forwardReferenceMatrix is None:
        instance._forwardReferenceInit()

    mem2=instance.nbytesReference

    return {
        'fastmat': {
            'func': func1,
            'args': args1,
            'Mem': mem1
        }, 'fm10': {
            'func': func1,
            'args': [args1[0], np.resize(args1[1], (numN, 10))]
        },  'numpy': {
            'func': func2,
            'args': args2,
            'Mem': mem2
        }, 'np10': {
            'func': func2,
            'args': [args1[0], np.resize(args1[1], (numN, 10))]
        }
    }


################################################## testInitForward()
def testInitForward(funcConstr, numSize, numN):
    instance=funcConstr(numSize)
    mem1=instance.nbytes

    func1=instance.forward
    func2=instance._forwardReference

    args=[arrTestDist((numN, 1), instance.dtype)]
    mem2=instance.nbytesReference

    return {
        'fastmat': {
            'func': func1,
            'args': args,
            'Mem': mem1
        }, 'fm10': {
            'func': func1,
            'args': [np.resize(args[0], (numN, 10))]
        }, 'fm100': {
            'func': func1,
            'args': [np.resize(args[0], (numN, 100))]
        }, 'numpy': {
            'func': func2,
            'args': args,
            'Mem': mem2
        }, 'np10': {
            'func': func2,
            'args': [np.resize(args[0], (numN, 10))]
        }
    }


################################################## testInitOverhead()
def testInitOverhead(funcConstr, numSize, numN):
    instance=funcConstr(numSize)
    mem=instance.nbytes

    funcF=instance.forward
    funcB=instance.backward

    M = 1

    args=[arrTestDist((numN, M), instance.dtype)]

    profFwd, profBwd = instance.profileForward, instance.profileBackward
    estFwd, estBwd = instance.estimateRuntime(M)

    return {
        'forward': {
            'func': funcF,
            'args': args,
            'Mem': mem,
            BENCH.RESULT_OVH_NESTED_F   : profFwd['overheadNested'],
            BENCH.RESULT_EFF_NESTED_F   : profFwd['effortNested'],
            BENCH.RESULT_COMPLEXITY_F   : profFwd['complexity'],
            BENCH.RESULT_ESTIMATE_FWD   : estFwd
        }, 'backward': {
            'func': funcB,
            'args': args,
            'Mem': mem,
            BENCH.RESULT_OVH_NESTED_B   : profBwd['overheadNested'],
            BENCH.RESULT_EFF_NESTED_B   : profBwd['effortNested'],
            BENCH.RESULT_COMPLEXITY_B   : profBwd['complexity'],
            BENCH.RESULT_ESTIMATE_BWD   : estBwd
        }
    }


################################################## testInitDtype()
def testInitDtype(funcConstr, numSize, numN):
    def generate(dtype):
        instance=funcConstr(numSize, dtype)

        # if no instance was created, tag this target as non-existant
        if instance is None:
            return None

        # either a Matrix instance or a call tuple structure is returned
        if isinstance(instance, Matrix):
            return {
                'func': instance.forward,
                'args': [arrTestDist((numN, 1), dtype)],
                'Mem': instance.nbytes
            }
        else:
            return {
                'func': instance[0],
                'args': instance[1],
                'Mem': instance[1][0].nbytes
            }

    # create instances for all datatypes listed
    result={name: generate(dtype)
            for dtype, name in NAME.TYPENAME.items() if dtype is not None}

    # prune None entries
    result={key: val for key, val in result.items() if (val is not None)}
    return result


################################################## testInitPerformance()
def testInitPerformance(funcConstr, numSize, numN):
    func, args=funcConstr(numSize)
    mem=args[0].nbytes

    return {
        'perf': {
            'func': func,
            'args': args,
            'Mem': mem
        }, 'perf10': {
            'func': func,
            'args': [args[0], np.resize(args[1],
                                        (args[1].shape[0], 10))] + args[2:]
        }
    }


################################################################################
################################################## class BenchmarkRunner
class Benchmark(Worker):

    def __init__(self, targetClass, **options):

        # extract options
        extraOptions = options.get('extraOptions', {})

        # by default, enable verbosity for issues and isolate problems
        self.cbStatus=self.printStatus

        # define defaults
        defaults={
            BENCH.COMMON: {
                # default naming of benchmark target
                BENCH.NAME          : dynFormat(
                    "%s.%s", NAME.CLASS, NAME.TARGET),
                BENCH.CAPTION       : dynFormat("%s", BENCH.NAME),

                # default benchmark abort criteria
                BENCH.LIMIT_MEMORY  : 1e6,  # kilobytes
                BENCH.LIMIT_INIT    : 1.,   # seconds
                BENCH.LIMIT_ITER    : 1.,  # seconds for all targets combined
                BENCH.LIMIT_SIZE    : 1e9,  # problem size

                # default benchmark parameters
                BENCH.STEP_START_K  : 1,
                BENCH.STEP_MINITEMS : 5,

                # default benchmark control functionals
                BENCH.FUNC_INIT     : None,
                BENCH.FUNC_GEN      : None,
                BENCH.FUNC_SIZE     : (lambda k: k),
                BENCH.FUNC_STEP     : (lambda k: k * 10 ** (1. / 12)),

                # default benchmark timing parameters
                BENCH.MEAS_MINTIME  : 0.003,
                BENCH.MEAS_MINREPS  : 7,
                BENCH.MEAS_MINRUNS  : 7,
                BENCH.MEAS_COMPACT  : True
            },
            BENCH.FORWARD: {
                BENCH.CAPTION       : dynFormat("%s (Multiplication)",
                                                BENCH.NAME),
                BENCH.FUNC_INIT     : testInitForward
            },
            BENCH.SOLVE: {
                BENCH.CAPTION       : dynFormat("%s (Solving a LSE)",
                                                BENCH.NAME),
                BENCH.FUNC_INIT     : testInitSolve
            },
            BENCH.OVERHEAD: {
                BENCH.CAPTION       : dynFormat("%s (call overhead)",
                                                BENCH.NAME),
                BENCH.FUNC_INIT     : testInitOverhead,
                BENCH.FUNC_SIZE     : (lambda k: 2 ** k),
                BENCH.FUNC_STEP     : (lambda k: k + 1)
            },
            BENCH.PERFORMANCE: {
                BENCH.CAPTION       : dynFormat("%s (Algorithm performance)",
                                                BENCH.NAME),
                BENCH.FUNC_INIT     : testInitPerformance
            },
            BENCH.DTYPES: {
                BENCH.CAPTION       : dynFormat("%s (impact of data type)",
                                                BENCH.NAME),
                BENCH.FUNC_INIT     : testInitDtype,
                BENCH.FUNC_SIZE     : (lambda k: 2 ** k),
                BENCH.FUNC_STEP     : (lambda k: k + 1)
            }
        }

        # call parent initialization with Test-specific options
        super(Benchmark, self).__init__(
            targetClass, targetOptionMethod='_getBenchmark',
            runnerDefaults=defaults, extraOptions=extraOptions)

    def getResult(self, nameResult, *nameParams):
        result=self.results.get(nameResult)
        header=result[BENCH.HEADER]
        indices=[header.index(key)
                 for key in (nameParams if len(nameParams) > 0 else header)]
        return result[BENCH.RESULT][:, indices]

    def _run(self, name, options):
        '''
        Perform a runtime performance evaluation.
        '''
        # output benchmark title
        self.emitStatus(fmtBold(">> %s" %(getattr(paramDict(options),
                                                  BENCH.CAPTION))))

        # perform the benchmarks
        run, result, header={}, {}, None
        result[BENCH.RESULT]=[]
        valK, numN=options[BENCH.STEP_START_K], 0
        numItems = 0
        while (run.get(BENCH.RESULT_MEMORY, 1) < options[BENCH.LIMIT_MEMORY] and
               run.get(BENCH.RESULT_INIT, 0.) < options[BENCH.LIMIT_INIT] and
               run.get(BENCH.RESULT_ITER, 0.) < options[BENCH.LIMIT_ITER] and
               run.get(BENCH.RESULT_SIZE, 1)  < options[BENCH.LIMIT_SIZE] or
               numItems < options[BENCH.STEP_MINITEMS]):
            # run test, init result dictionary for this run
            run.clear()

            # estimate next size
            valK, lastValK=options[BENCH.FUNC_STEP](valK), valK
            if valK == lastValK:
                raise ValueError(
                    "Stepping function caused endless loop: abort. [%s]" %(
                        options))
            numK=int(round(valK))
            numN, lastNumN=options[BENCH.FUNC_SIZE](numK), numN

            # if problem size did not change since last iter, skip
            if lastNumN == numN:
                #print("skipping %f (%d, %d)"%(valK, lastNumN, numN))
                continue
            run[BENCH.RESULT_SIZE] = numN

            # count number of elements in result to force minimal amount of
            numItems += 1

            ####################### construct test instance and init call list
            calls={}
            run[BENCH.RESULT_INIT]=timeit(
                lambda : calls.update(options[BENCH.FUNC_INIT](
                    options[BENCH.FUNC_GEN], numK, numN)),
                number=1)

            # determine total memory consumption of all call-related stuff and
            # matrix memory consumption of each call target separately
            totalMemUsed=getMemoryFootprint(calls) / 1e3
            for tag, call in calls.items():
                # memory consumption of each calls' matrix instance
                if 'Mem' in call:
                    memCall=call['Mem'] / 1e3
                    totalMemUsed += memCall
                    run["%sMem" %(tag)]=memCall

                # copy effort columns if defined in testInitCalls
                for key in [BENCH.RESULT_OVH_NESTED_F,
                            BENCH.RESULT_OVH_NESTED_B,
                            BENCH.RESULT_EFF_NESTED_F,
                            BENCH.RESULT_EFF_NESTED_B,
                            BENCH.RESULT_COMPLEXITY_F,
                            BENCH.RESULT_COMPLEXITY_B,
                            BENCH.RESULT_ESTIMATE_FWD,
                            BENCH.RESULT_ESTIMATE_BWD]:
                    if key in call:
                        run[key] = call[key]

            run[BENCH.RESULT_MEMORY]=totalMemUsed

            # do one extra warm-up run before starting actual measurements
            # to fetch and all required code and fetch data memory
            timeCalls(calls, **options)

            # determine runtime statistics and compose runs
            # alongside, check timing limits
#            run[BENCH.RESULT_ITER] = 0
            iterTime = time.time()
            run.update(timeCalls(calls, **options))
#            for tag in calls.keys():
#                run[BENCH.RESULT_ITER] += run["%sAvg" %(tag)]
            run[BENCH.RESULT_ITER] = time.time() - iterTime

            result[BENCH.RESULT].append(run.copy())

            # format instant output
            def formatToken(token):
                try:
                    # int or scientific notation
                    return ("%10d" if isinstance(token, int)
                            else "%10.3e") %(token)
                except TypeError:
                    # if that fails: make a string out of whatever it is
                    return "%11s" %(token)

            # print header in first iteration
            if BENCH.HEADER not in result:
                # determine
                keys=set([key
                          for step in result[BENCH.RESULT] for key in step])

                # let some key names have priority in ordering of header.
                priority=[BENCH.RESULT_SIZE, BENCH.RESULT_MEMORY,
                          BENCH.RESULT_INIT, BENCH.RESULT_ITER]

                # return the priority keys (in order of specification),
                # followed by the other keys (alphabetically sorted)
                result[BENCH.HEADER]=(
                    [key for key in priority if key in keys] +
                    [key for key in sorted(keys) if key not in priority])

                # emit header status line (--> printing)
                strHeader=" ".join([formatToken(f)
                                    for f in result[BENCH.HEADER]])
                self.emitStatus(strHeader + '\n' + '-' * len(strHeader))

            # output runs in same order as header
            row=['.'] * len(result[BENCH.HEADER])
            for key in run.keys():
                try:
                    row[result[BENCH.HEADER].index(key)]=run[key]
                except KeyError:
                    pass

            # print the whole stuff
            self.emitStatus("[%s]" %(", ".join([formatToken(f) for f in row])))

        # collate result dictionary structure to numpy array
        res, header=result[BENCH.RESULT], result[BENCH.HEADER]
        arrRes=np.zeros((len(res), len(header)), dtype=np.float64)
        for ii, item in enumerate(res):
            arrRes[ii, :]=np.array([item.get(key, np.NaN)
                                    for key in header])
        result[BENCH.RESULT]=arrRes

        return result

    # control verbosity of _run() and printStatus()
    @property
    def verbosity(self):
        return self.cbStatus == self.printStatus

    @verbosity.setter
    def verbosity(self, value):
        self.cbStatus=self.printStatus if value is True else None

    def printStatus(self, message):
        print(message)

    def getFilename(self, nameResult, path='', addVersionTag=True):
        return os.path.join(path, "%s.%s%s.csv" %(
            self.target.__class__.__name__, nameResult,
            ".%s" %(__version__) if addVersionTag else ""))

    def saveResult(self, nameResult, outPath='', addVersionTag=True,
                   strFormat='%.6e', chrDelimiter=',', chrNewline=os.linesep):
        '''
        Print data table to csv file. The result is organized as a list of
        dicts containing information in key=value pairs. The header will be
        extracted from the keys of the result data dictionaries. If the file
        does not exist it will be created.
        '''
        if nameResult not in self.results:
            raise ValueError("No benchmark result of such name found")

        result=self.results[nameResult]
        arrResult=result[BENCH.RESULT]
        lstHeader=result[BENCH.HEADER]

        filename=self.getFilename(nameResult, path=outPath,
                                  addVersionTag=addVersionTag)
        # force existance of output file
        dirname=os.path.dirname(filename)
        if dirname != '' and not os.path.exists(dirname):
            os.makedirs(dirname)

        # save csv-file to memory
        np.savetxt(filename, arrResult,
                   header=chrDelimiter.join(lstHeader), comments='',
                   fmt=strFormat, newline=chrNewline, delimiter=chrDelimiter)

        return filename



# -*- coding: utf-8 -*-
'''
  demo/matrixSAFT.py
 -------------------------------------------------- part of the fastmat demos

  Demonstration on how to use fastMat for SAFT.


  Author      : wcw
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import sys
from timeit import timeit

# import numpy functionality
import numpy as np
import scipy as sp

################################################## import modules
try:
    from matplotlib.pyplot import *               # plotting
except ImportError:
    print("matplotlib not found. Please consider installing it to proceed.")
    sys.exit(0)

try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat


################################################## import cython-optimized core
import pyximport; pyximport.install()
from matrixSAFTmask import SaftMaskCore


###############################################################################
# SAFT demo
###############################################################################
print("fastmat demo: Synthetic Aperture Focussing Technique (SAFT)")
print("-----------------------------------------------------------")

# x - Parameter für die Matrix (Unterscheidung)
# a - Rekovektor
# b -  Messvektor
# s_x - Spezifisches X
# j - Zeilenindex Matrix - Zeitrichtung (numN)
# dx - Schrittweite in Breite (M) (numD)
# dz - Schrittweite in Tiefe (N) (numS)


refData = sp.io.loadmat('datenSAFT.mat')
dX = refData['dx'] * 1000
dZ = refData['dz'] * 1000
numN        = 2040
numK        = 50
numD        = 1000 if numN == 2040 else numN / 2

for key, value in {
    'Samples in time dimension (numN)': numN,
    'Samples in spatial dimension (numD)': numD,
    'Summation window width (numK)': numK}.items():
    print("%40s = %g" %(key, value))



##################################################  Generate SAFT aperture masks
print("\nGenerating aperture masks and SAFT matrix:")

# create an array to store index mapping
# (row-index [:, 0] to col-index[:, 1], data is [:, 2])
arrI = np.zeros((numN, 3), dtype=np.int32)
arrI[:, 0] = np.arange(numN)
arrI[:, 2] = np.ones(numN, dtype=np.int32)

sizeFull = numN * numD
sizeItem = numN
cntBlocks = sizeFull // sizeItem

# Generate sparse matrices: iterate over numK
matSparse = []
for kk in range(0, numK):
    # determine row-to-col index mapping for kk
    arrI[:, 1] = np.sqrt((kk * dX) ** 2 + (arrI[:, 0] * dZ) ** 2) / dZ + 0.5

    # select all entries within matrix range
    selection = np.where((arrI[:, 1] < sizeItem) * (arrI[:, 1] >= 0))[0]
    mapping = arrI[selection].astype(np.int16)

    # stop when submatrices are empty
    if selection.size < 1:
        break

    matSparse.append(
        sp.sparse.coo_matrix(
                    (mapping[:, 2], (mapping[:, 0], mapping[:, 1])),
                    shape=(sizeItem, sizeItem), dtype=np.int8))
fastmatSparse = [fastmat.Sparse(mat.tocsc()) for mat in matSparse]
cntK = len(matSparse)



################################################## user-defined SAFT class
class SaftMaskClass(fastmat.Matrix):
    def __init__(self, numBlocks, sizeItem, *masks):
        if any((mask.ndim != 2 or mask.shape[1] != 2) for mask in masks):
            raise ValueError("Masks of SaftMask must be of shape Nx2!")
        self._masks = tuple(mask.astype(np.uint16, copy=True, order='C')
                            for mask in masks)
        self._sizeItem = sizeItem
        self._numBlocks = numBlocks
        self._numMasks = len(masks)
        numN = numBlocks * self._sizeItem
        self._initProperties(numN, numN, np.int8)

    def _forward(self, arrX):
        return SaftMaskCore(arrX, self._sizeItem, self._numBlocks, self._masks,
                            False)

    def _backward(self, arrX):
        return SaftMaskCore(arrX, self._sizeItem, self._numBlocks, self._masks,
                            True)

################################################## Use special fastmat class
print(" %-20s : %s" %("'matSaftMaskClass'",
                      "user-defined fastmat class, cython-optimized"))
matSaftMaskClass = SaftMaskClass(numD, sizeItem,
                                 *(tuple(np.vstack((mat.row, mat.col)).T
                                         for mat in matSparse)))
print(" " * 24 + repr(matSaftMaskClass))


figure(1)
rc('image', aspect='auto', interpolation='none')

arrScanData = refData['data'].astype(np.float32)
#print(arrScanData.min(), arrScanData.max(), arrScanData.dtype)
#vecInput = arrScanData.reshape((-1, 1)).astype(np.float32)
vecInput = arrScanData.T.reshape((-1, 1))

vecOutput = None
def doIt():
    global vecOutput
    vecOutput = matSaftMaskClass * vecInput

cnt = 5
doIt()
print(" %20s = %12.4e s (%d repetitions)" %(
    "runtime", timeit(doIt, number=cnt) / cnt, cnt))

arrReconstruction = vecOutput.T.reshape(arrScanData.T.shape).T

subplot(1, 3, 1)
imshow(arrScanData)

subplot(1, 12, 5)
imshow(vecInput)
subplot(1, 12, 6)
imshow(vecOutput)

subplot(1, 2, 2)
imshow(arrReconstruction)

show()



# -*- coding: utf-8 -*-
'''
  fastmat/core/cmath.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for fastmat type handling support.

  Author      : wcw
  Introduced  : 2016-11-16
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
import numpy as np
cimport numpy as np

################################################## define unique type-id types

ctypedef np.uint8_t ftype
ctypedef np.uint8_t nptype
ctypedef np.npy_intp intsize

################################################## define global fused types
#
# fastmat-wide definitions of fused types which specify the datatypes allowed
# in memoryviews of ndarrays.
#
# Fused types work in a way that each argument of each fused type used as
# parameters must be of the same explicit type. i.e. if a function accepts
# two parameters of the same-named fused type (allowing float and double),
# then both of these parameters must either be float or double and no combina-
# tions between these types are allowed. To allow combinations, each parameter
# must have its own fused type, even if they contain the same definitions.
# Further, these types also must not be derived from other fused types.

# WARNING: be sure to keep the NUM_TYPES definition in types.h correct!
cdef extern from "types.h":
    enum: NUM_TYPES

ctypedef enum TYPES:
    TYPE_INT8
    TYPE_INT32
    TYPE_INT64
    TYPE_FLOAT32
    TYPE_FLOAT64
    TYPE_COMPLEX64
    TYPE_COMPLEX128
    TYPE_INVALID
    TYPE_NUM

################################ fused type classes

ctypedef fused TYPE_INT:
    np.int8_t
    np.int32_t
    np.int64_t

ctypedef fused TYPE_REAL:
    np.float32_t
    np.float64_t

ctypedef fused TYPE_COMPLEX:
    np.complex64_t
    np.complex128_t

ctypedef fused TYPE_FLOAT:
    np.float32_t
    np.float64_t
    np.complex64_t
    np.complex128_t

ctypedef fused TYPE_ALL:
    np.int8_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t
    np.complex64_t
    np.complex128_t


################################ one round of fused types for input arguments
ctypedef fused TYPE_IN:
    np.int8_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t
    np.complex64_t
    np.complex128_t

ctypedef fused TYPE_IN_R:
    np.int8_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t

ctypedef fused TYPE_IN_I:
    np.int8_t
    np.int32_t
    np.int64_t


################################ one round of fused types for operand arguments
ctypedef fused TYPE_OP:
    np.int8_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t
    np.complex64_t
    np.complex128_t

ctypedef fused TYPE_OP_R:
    np.int8_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t

ctypedef fused TYPE_OP_I:
    np.int8_t
    np.int32_t
    np.int64_t


################################ type information structures

ctypedef struct INFO_TYPE_s:
    nptype          typeNum
    ftype           fusedType
    ftype           *promote
    np.float64_t    eps
    np.float64_t    min
    np.float64_t    max
    int             dsize
    bint            isComplex
    bint            isInt
    bint            isNumber


ctypedef struct INFO_ARR_s:
    INFO_TYPE_s     *dtype
    int             nDim
    intsize         numN
    intsize         numM


################################################## type handling

cdef INFO_TYPE_s typeInfo[NUM_TYPES]
cdef ftype typeSelection[<int> np.NPY_NTYPES]

cdef INFO_TYPE_s *_getTypeInfo(object)
cdef nptype _getNpType(np.ndarray arr)
cdef ftype _getFType(np.ndarray arr)

cpdef np.float64_t _getTypeEps(object dtype)
cpdef np.float64_t _getTypeMin(object dtype)
cpdef np.float64_t _getTypeMax(object dtype)


################################################## type Promotion stuff

cdef ftype _promoteTypeNums(ftype, ftype)
cpdef object safeTypeExpansion(object)
